package ru.microem.readerlibrary;

import android.Manifest;
import android.app.Activity;
import android.app.PendingIntent;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattDescriptor;
import android.bluetooth.BluetoothManager;
import android.bluetooth.BluetoothProfile;
import android.bluetooth.le.BluetoothLeScanner;
import android.bluetooth.le.ScanCallback;
import android.bluetooth.le.ScanFilter;
import android.bluetooth.le.ScanResult;
import android.bluetooth.le.ScanSettings;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbDeviceConnection;
import android.hardware.usb.UsbEndpoint;
import android.hardware.usb.UsbInterface;
import android.hardware.usb.UsbManager;
import android.os.Build;
import android.os.Handler;
import android.util.Log;
import android.widget.Toast;


import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Random;
import java.util.UUID;
import java.util.Vector;

import static android.bluetooth.BluetoothDevice.PHY_LE_1M_MASK;
import static android.bluetooth.BluetoothDevice.PHY_LE_CODED_MASK;
import static android.bluetooth.BluetoothDevice.TRANSPORT_LE;
import static android.content.ContentValues.TAG;
import static android.content.Context.BLUETOOTH_SERVICE;

public class UemReader {
    public static final String UEM_INTERFACE_UNKNOWN = "Unknown Interface";
    public static final String UEM_INTERFACE_USB = "USB Interface";
    public static final String UEM_INTERFACE_BLUETOOTH = "Bluetooth Interface";

    public static final String UEM_STATUS_SUCCESS = "OK";
    public static final String UEM_STATUS_OPERATION_PENDING = "Operation in progress";
    public static final String UEM_ERROR_UNSUPPORTED_FEATURE = "Feature not supported";
    public static final String UEM_ERROR_COMMUNICATION_DATA_LOST = "Communication data lost";
    public static final String UEM_ERROR_INCORRECT_PARAMETER = "Incorrect parameter";
    public static final String UEM_ERROR_UNEXPECTED = "Unexpected error";
    public static final String UEM_ERROR_ACCESS = "Access error";
    public static final String UEM_ERROR_NOT_TRANSACTED = "Not transacted";
    public static final String UEM_ERROR_READER_INCORRECT_NAME = "Incorrect reader name";
    public static final String UEM_ERROR_READER_CONNECTION_FAILED = "Failed to connect to the reader";
    public static final String UEM_ERROR_READER_NOT_CONNECTED = "Reader not connected";
    public static final String UEM_ERROR_READER_ALREADY_CONNECTED = "Reader already connected";
    public static final String UEM_ERROR_READER_INCORRECT_RESPONSE = "Incorrect reader response";
    public static final String UEM_ERROR_READER_RESPONSE_FAILURE = "Reader not responding";
    public static final String UEM_ERROR_READER_UNSUCCESSFUL = "Reader returned error code";
    public static final String UEM_ERROR_SAM_APDU = "SAM: APDU error";
    public static final String UEM_ERROR_SAM_INVALID_MAC = "SAM: Invalid MAC";
    public static final String UEM_ERROR_SAM_AUTHENTICATION_FAILED = "SAM: Authentication failed";

    private Context mContext = null;

    //private HashMap<Object, String> mConnectedReaders = new HashMap<Object, String>();
    private String mInterfaceType = null;

    private String mLastError = UEM_STATUS_SUCCESS;

    private byte mLastErrorCode = 0x00;

    private Object mConnection = null;

    public Object debugData = "";

    public UemReader(Context context, boolean fStartReceivingBroadcasts) {
        initialize(context, fStartReceivingBroadcasts);
    }

    public UemReader(Context context) {
        initialize(context, true);
    }

    private void initialize(Context context, boolean fStartReceivingBroadcasts) {
        if (context == null)
            throw new NullPointerException("Parameter is null!");
        this.mContext = context;

        this.USB.mReader = this;
        this.Bluetooth.mReader = this;

        this.USB.initialize();

        if (fStartReceivingBroadcasts)
            this.startReceivingBroadcasts();
    }

    public String getLastError() {
        return mLastError;
    }

    public byte getLastErrorCode() {
        return mLastErrorCode;
    }

    public class CommandsReader {
        public static final int UEM_COLOR_RED = 0x01;
        public static final int UEM_COLOR_GREEN = 0x02;
        public static final int UEM_COLOR_BLUE = 0x04;

        public static final int UEM_RF_MODE_ISO14443_TRANSMIT_BAUDRATE_106 = 0x00;
        public static final int UEM_RF_MODE_ISO14443_TRANSMIT_BAUDRATE_212 = 0x04;
        public static final int UEM_RF_MODE_ISO14443_TRANSMIT_BAUDRATE_424 = 0x08;
        public static final int UEM_RF_MODE_ISO14443_TRANSMIT_BAUDRATE_848 = 0x0C;

        public static final int UEM_RF_MODE_ISO14443_RECEIVE_BAUDRATE_106 = 0x00;
        public static final int UEM_RF_MODE_ISO14443_RECEIVE_BAUDRATE_212 = 0x01;
        public static final int UEM_RF_MODE_ISO14443_RECEIVE_BAUDRATE_424 = 0x02;
        public static final int UEM_RF_MODE_ISO14443_RECEIVE_BAUDRATE_848 = 0x03;

        public byte[] directIOReader(byte[] command) {
            return dispatchCommand(command);
        }

        public boolean beep(int beepCount) {
            dispatchCommand(new byte[]{0x05, (byte)beepCount});
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public boolean blink(int blinkColor, int blinkCount, int postColor) {
            dispatchCommand(new byte[]{0x07, (byte)blinkColor, (byte)blinkCount, (byte)postColor});
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public boolean switchOnRadioController() {
            dispatchCommand(new byte[]{0x10});
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public boolean switchOffRadioController() {
            dispatchCommand(new byte[]{0x04, (byte)0x80, 0x01});
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public boolean setRadioMode(int mode) {
            dispatchCommand(new byte[]{0x51, (byte)mode});
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public boolean toggleRadioOffOn(int delayMs) {
            dispatchCommand(new byte[]{0x20, (byte)(delayMs & 0xFF), (byte)((delayMs & 0xFFFF) >> 8)});
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public byte[] getState() {
            return dispatchCommand(new byte[]{0x0E});
        }

        public byte[] getVersion() {
            return dispatchCommand(new byte[]{0x64});
        }

        public byte[] getSerial() {
            return dispatchCommand(new byte[]{0x22});
        }

        public byte[] getReaderID() {
            return dispatchCommand(new byte[]{0x58, 0x58});
        }

        public boolean eraseFlash() {
            dispatchCommand(new byte[]{0x70, 0x21, 0x43, 0x65, (byte)0x87});
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public boolean writeFlashValue(int address, byte[] value) {
            if (value == null || value.length != 16) {
                mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                return false;
            }
            ByteArrayOutputStream commandReader = new ByteArrayOutputStream();
            commandReader.write(0x79);
            commandReader.write((byte)(address & 0xFF));
            commandReader.write((byte)((address & 0xFFFF) >> 8));
            commandReader.write(value, 0, value.length);
            byte[] res = dispatchCommand(commandReader.toByteArray());
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public boolean resetReaderAuthentication() {
            dispatchCommand(new byte[]{0x6F, (byte)0x00, (byte)0x00, (byte)0x00});
            boolean res = getLastError().equals(UEM_STATUS_SUCCESS);
            if (res)
                encryptedMode = false;
            return res;
        }

        private byte[] sessionIV = new byte[16];
        private byte[] sessionKey = new byte[16];
        private boolean encryptedMode = false;
        public boolean authenticateReader(int keyFlashAddress, byte[] key) {
            byte[] rndA = new byte[16];
            byte[] rndB = null;

            Arrays.fill(sessionIV, (byte)0x00);

            //new byte[] {0x56, 0x38, (byte)0x87, 0x23, 0x56, 0x38, (byte)0x87, 0x23, 0x56, 0x38, (byte)0x87, 0x23, 0x56, 0x38, (byte)0x87, 0x23};
            byte[] res = dispatchCommand(new byte[]{0x6F, (byte)0xAA, (byte)(keyFlashAddress & 0xFF), (byte)((keyFlashAddress >> 8) & 0xFF)});

            if (!getLastError().equals(UEM_STATUS_SUCCESS))
                return false;

            if (res == null || res.length == 0) {
                mLastError = UEM_ERROR_UNEXPECTED;
                return false;
            }

            (new Random()).nextBytes(rndA);

            byte[] out = new byte[33];

            System.arraycopy(rndA, 0, out, 0, 16);

            rndB = Helper.AES.aes128DecryptCbc(res, 16, key, new byte[16], null);

            System.arraycopy(rndB, 1, out, 16, 15);

            out[16 + 15] = rndB[0];

            byte[] tmp = Helper.AES.aes128EncryptCbc(out, 32, key, Helper.AES.aes128DecryptCbcIvResult, null);

            ByteArrayOutputStream commandReader = new ByteArrayOutputStream();
            commandReader.write(0x6F);
            commandReader.write(0xAF);
            commandReader.write((byte) (keyFlashAddress & 0xFF));
            commandReader.write((byte) ((keyFlashAddress & 0xFFFF) >> 8));
            commandReader.write(tmp, 0, tmp.length);
            res = dispatchCommand(commandReader.toByteArray());

            if (!getLastError().equals(UEM_STATUS_SUCCESS))
                return false;

            byte[] rndAs = Helper.AES.aes128DecryptCbc(res, 16, key, Helper.AES.aes128EncryptCbcIvResult, null);

            byte[] rndAtest = new byte[16];

            rndAtest[0] = rndAs[15];

            System.arraycopy(rndAs, 0, rndAtest, 1, 15);

            if (!Arrays.equals(rndA, rndAtest)) {
                mLastError = UEM_ERROR_SAM_AUTHENTICATION_FAILED;
                return false;
            }

            System.arraycopy(rndA, 0, sessionKey, 0, 4);
            System.arraycopy(rndB, 0, sessionKey, 4, 4);
            System.arraycopy(rndA, 12, sessionKey, 8, 4);
            System.arraycopy(rndB, 12, sessionKey, 12, 4);

            Arrays.fill(sessionIV, (byte)0x00);

            System.arraycopy(sessionKey, 12, sessionKey, 12, 4);

            encryptedMode = true;
            return true;
        }

        public boolean getEncryptionMode() {
            return encryptedMode;
        }

        public void setEncryptionMode(boolean encryption) {
            encryptedMode = encryption;
        }
    }

    public CommandsReader Reader = new CommandsReader();

    public class CommandsSAM {
        public static final int UEM_SAM_HOST_MODE_PLAIN = 0x00;
        public static final int UEM_SAM_HOST_MODE_MAC = 0x01;
        public static final int UEM_SAM_HOST_MODE_FULL = 0x02;
        public static final int UEM_SAM_KEY_KIND_TDEA_DES_FIRE = 0x00;
        public static final int UEM_SAM_KEY_KIND_TDEA_ISO10116_CRC16 = 0x01;
        public static final int UEM_SAM_KEY_KIND_MIFARE = 0x02;
        public static final int UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY = 0x03;
        public static final int UEM_SAM_KEY_KIND_AES128 = 0x04;
        public static final int UEM_SAM_KEY_KIND_AES192 = 0x05;
        public static final int UEM_SAM_KEY_KIND_TDEA_ISO10116_CRC32 = 0x06;

        public byte[] ATR = null;
        public byte[] version = null;
        public byte[] SW = new byte[] {(byte)0x90, 0x00};

        public int getCurrentHolder() {
            byte[] res = dispatchCommand(new byte[]{(byte)0xB4});
            if (!getLastError().equals(UEM_STATUS_SUCCESS) || res == null)
                return -1;
            return res[0];
        }

        public byte[] setCurrentHolder(int holder) {
            byte[] res = dispatchCommand(new byte[]{(byte)0xB2, (byte)(0x00000010 | holder & 0x000000FF)});
            if (!getLastError().equals(UEM_STATUS_SUCCESS) || res == null) {
                ATR = null;
                return ATR;
            }
            ATR = Arrays.copyOfRange(res, 2, res.length);
            return ATR;
        }

        public byte[] transceiveAPDU(byte[] APDU) {
            if (APDU == null || APDU.length == 0) {
                mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                return null;
            }
            SW[0] = 0x00;
            SW[1] = 0x00;
            ByteArrayOutputStream commandReader = new ByteArrayOutputStream();
            commandReader.write(0xB6);
            commandReader.write((byte)(APDU.length & 0xFF));
            commandReader.write((byte)((APDU.length & 0xFFFF) >> 8));
            commandReader.write(APDU, 0, APDU.length);
            byte[] res = dispatchCommand(commandReader.toByteArray());
            int len = 0;
            if (getLastErrorCode() == (byte)0xC8) {//UEM_SAM_APDU_ERR
                if (res == null || res.length < 4) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }
                SW[0] = res[res.length - 2];
                SW[1] = res[res.length - 1];
                mLastError = UEM_ERROR_SAM_APDU;
                len = res[0] + res[1] * 256;
                return Arrays.copyOfRange(res, 2, len);
            } else if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                return null;
            }
            if (res == null || res.length < 4) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return null;
            }
            SW[0] = res[res.length - 2];
            SW[1] = res[res.length - 1];
            len = res[0] + res[1] * 256;
            return Arrays.copyOfRange(res, 2, len);
        }

        public byte[] getVersion(int logicalChannel) {
            return transceiveAPDU(new byte[]{(byte)(0x80 | logicalChannel), 0x60, 0x00, 0x00, 0x00});
        }

        public boolean resetCold() {
            dispatchCommand(new byte[]{(byte)0xBA, 0x21, 0x43, 0x65, (byte)0x87});
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public boolean killCardAuthentication() {
            dispatchCommand(new byte[]{(byte)0xBA, 0x01, 0x43, 0x65, (byte)0x87});
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public boolean killTotalAuthentication() {
            dispatchCommand(new byte[]{(byte)0xBA, 0x00, 0x43, 0x65, (byte)0x87});
            return getLastError().equals(UEM_STATUS_SUCCESS);
        }

        public boolean authenticateHost(int logicalChannel, int hostMode, int keyKind, int keyNumber, int keyVersion, byte[] key) {
            if (key == null || key.length == 0) {
                mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                return false;
            }
            return samAuthenticateHost(logicalChannel, keyKind, keyNumber, keyVersion, key, hostMode);
        }

        public boolean lock(int logicalChannel, int keyKind, int keyNumber, int keyVersion, byte[] unlockKey, int unlockKeyNumber, int unlockKeyVersion) {
            if (unlockKey != null && unlockKey.length == 0) {
                mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                return false;
            }
            return samLockUnlock(logicalChannel, (byte)((unlockKey == null) ? 0x01 : 0x02), keyKind, keyNumber, keyVersion, unlockKey, unlockKeyNumber, unlockKeyVersion, 0);
        }

        public boolean unlock(int logicalChannel, int keyKind, int keyNumber, int keyVersion, byte[] unlockKey) {
            if (unlockKey != null && unlockKey.length == 0) {
                mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                return false;
            }
            return samLockUnlock(logicalChannel, (byte)0x00, keyKind, keyNumber, keyVersion, unlockKey, (byte)0x00, (byte)0x00, 0);
        }

        private final int SAM_BUF_SIZE = 256;
        private byte[] mSAM_Plain = new byte[SAM_BUF_SIZE];
        private byte[] mSAM_Coded = new byte[SAM_BUF_SIZE];
        private byte[] mSessionKeyEnc = new byte[16];
        private byte[] mSessionKeyMAC = new byte[16];

        //private final char MI_WRONG_PARAMETER_VALUE = 0xC4;
        private byte[] mKeyX = new byte[24];

        private boolean samAuthenticateHost(int LogicalChannel,
                                            int KeyKind,
                                            int KeyNo,
                                            int KeyVer,
                                            byte[] Kx,
                                            int HostMode )
        {
            int i, j;
            int SendLen = 0;
            byte[] ToSM;
            byte[] From;
            byte[] SV1 = new byte[16];
            byte[] SV2 = new byte[16];
            byte[] MAC16 = new byte[16];
            byte[] MAC24 = new byte[24];
            byte[] Kxe = new byte[24];
            char status = 0;

            // Rnd1 p.6 to SAM
            byte[] Rnd1 = new byte[]{0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B};
            (new Random()).nextBytes(Rnd1);

            // RndA p.14 to SAM
            byte[] RndA = new byte[]{0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F};
            (new Random()).nextBytes(RndA);

            ToSM = mSAM_Plain;
            From = mSAM_Coded;
            ToSM[SendLen++] = (byte)((byte)0x80 | (byte)LogicalChannel);
            ToSM[SendLen++] = (byte)0xa4;
            ToSM[SendLen++] = 0x00; // P1
            ToSM[SendLen++] = 0x00; // P2
            ToSM[SendLen++] = 0x03; // Lc
            ToSM[SendLen++] = (byte)KeyNo;
            ToSM[SendLen++] = (byte)KeyVer;
            ToSM[SendLen++] = (byte)HostMode;
            ToSM[SendLen++] = 0x00;    // Le

            From = transceiveAPDU(Arrays.copyOf(ToSM, SendLen));

            if( From == null )
                return false;
            if( ((SW[0] & 0xFF) != 0x90) || ((SW[1] & 0xFF) != 0xAF) ) {
                mLastError = UEM_ERROR_SAM_APDU;
                return false;
            }
            if( From.length == 0 ) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return false;
            }

            for(i = 0; i < 5; i++) {
                SV1[i] = Rnd1[7 + i];
                SV1[5 + i] = From[7 + i];
                SV1[10 + i] = (byte)(From[i] ^ Rnd1[i]);
            }
            SV1[15] = (byte)0x91;

            switch(KeyKind) {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    Kxe = Helper.AES.aes192Encrypt24( SV1, Kx );
                    break;
                default:
                    Kxe = Helper.AES.aes128Encrypt16( SV1, Kx );
            }

            byte[] MacData = new byte[From.length + 4];
            System.arraycopy(From, 0, MacData, 0, From.length);
            MacData[From.length] = (byte)HostMode;
            MacData[From.length + 1] = 0;
            MacData[From.length + 2] = 0;
            MacData[From.length + 3] = 0;

            switch(KeyKind) {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    MAC24 = Helper.AES.aes192CbcEnMac( MacData, 24, Kx, MAC24 );
                    break;
                default:
                    MAC16 = Helper.AES.aes128CbcEnMAC( MacData, 16, Kx, MAC16 );
            }

            // p.7
            SendLen = 0;
            ToSM = mSAM_Plain;
            ToSM[SendLen++] = (byte)(0x80 | LogicalChannel);
            ToSM[SendLen++] = (byte)0xa4;
            ToSM[SendLen++] = 0x00; // P1
            ToSM[SendLen++] = 0x00; // P2
            ToSM[SendLen++] = 0x14; // Lc
            switch(KeyKind) {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    for( i = 0, j = 1; i < 12; i++, j+=2 )
                        ToSM[SendLen++] = MAC24[j];
                    break;
                default:
                    for( i = 0, j = 1; i < 8; i++, j+=2 )
                        ToSM[SendLen++] = MAC16[j];
            }
            System.arraycopy(Rnd1, 0, ToSM, SendLen, Rnd1.length);
            SendLen += Rnd1.length;
            ToSM[SendLen++] = 0x00;    // Le

            From = transceiveAPDU(Arrays.copyOf(ToSM, SendLen));

            if( From == null )
                return false;
            if( ((SW[0] & 0xFF) != 0x90) || ((SW[1] & 0xFF) != 0xAF) ) {
                mLastError = UEM_ERROR_SAM_APDU;
                return false;
            }
            if( From.length == 0 ) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return false;
            }

            byte[] TmpData = new byte[16];
            System.arraycopy(Rnd1, 0, TmpData, 0, Rnd1.length);
            TmpData[Rnd1.length] = (byte)HostMode;
            TmpData[Rnd1.length + 1] = 0;
            TmpData[Rnd1.length + 2] = 0;
            TmpData[Rnd1.length + 3] = 0;

            byte[] TestMAC = new byte[8];
            switch(KeyKind) {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    MAC24 = Helper.AES.aes192CbcEnMac( TmpData, 24, Kx, MAC24 );

                    for( i = 0, j = 1; i < 8; i++, j+=3 )
                        TestMAC[i] = MAC24[j];

                    break;
                default:
                    MAC16 = Helper.AES.aes128CbcEnMAC( TmpData, 16, Kx, MAC16 );

                    for( i = 0, j = 1; i < 8; i++, j+=2 )
                        TestMAC[i] = MAC16[j];
            }

            if( !Arrays.equals(TestMAC, Arrays.copyOf(From, TestMAC.length)) ) {
                mLastError = UEM_ERROR_SAM_INVALID_MAC;
                return false;
            }

            byte[] RndB;
            switch(KeyKind) {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    RndB = Helper.AES.aes192Decrypt24( Arrays.copyOfRange(From, 16, From.length), Kxe );
                    break;
                default:
                    RndB = Helper.AES.aes128Decrypt16( Arrays.copyOfRange(From, 8, From.length), Kxe );
            }

            byte[] RndARndBB = new byte[RndA.length + RndB.length];
            System.arraycopy(RndA, 0, RndARndBB, 0, RndA.length);
            System.arraycopy(RndB, 2, RndARndBB, RndA.length, RndB.length - 2);
            System.arraycopy(RndB, 0, RndARndBB, RndA.length + RndB.length - 2, 2);

            SendLen = 0;
            ToSM = mSAM_Plain;
            ToSM[SendLen++] = (byte)(0x80 | LogicalChannel);
            ToSM[SendLen++] = (byte)0xa4;
            ToSM[SendLen++] = 0x00; // P1
            ToSM[SendLen++] = 0x00; // P2
            ToSM[SendLen++] = 0x20; // Lc
            byte[] EncRndARndBB = null;
            switch(KeyKind)
            {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    byte[] IV24 = new byte[24];
                    Arrays.fill(IV24, (byte)0x00);
                    EncRndARndBB = Helper.AES.aes192EncryptCbc(
                            RndARndBB,
                            RndARndBB.length,
                            Kxe,
                            IV24,
                            null );
                    break;
                default:
                    byte[] IV16 = new byte[16];
                    Arrays.fill(IV16, (byte)0x00);
                    EncRndARndBB = Helper.AES.aes128EncryptCbc(
                            RndARndBB,
                            RndARndBB.length,
                            Kxe,
                            IV16,
                            null );
            }
            System.arraycopy(EncRndARndBB, 0, ToSM, SendLen, EncRndARndBB.length);
            SendLen += EncRndARndBB.length;
            ToSM[SendLen++] = 0x00;    // Le

            From = transceiveAPDU(Arrays.copyOf(ToSM, SendLen));

            if( From == null )
                return false;
            if( ((SW[0] & 0xFF) != 0x90) || ((SW[1] & 0xFF) != 0x00) ) {
                mLastError = UEM_ERROR_SAM_APDU;
                return false;
            }
            if( From.length == 0 ) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return false;
            }

            byte[] TestRndAA = null;
            switch(KeyKind)
            {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    TestRndAA = Helper.AES.aes192Decrypt24( From, Kxe );
                    break;
                default:
                    TestRndAA = Helper.AES.aes128Decrypt16( From, Kxe );
            }

            byte[] RndAA = new byte[RndA.length];
            System.arraycopy(RndA, 2, RndAA, 0, RndAA.length - 2);
            System.arraycopy(RndA, 0, RndAA, RndAA.length - 2, 2);

            if( !Arrays.equals(RndAA, TestRndAA) ) {
                mLastError = UEM_ERROR_SAM_AUTHENTICATION_FAILED;
                return false;
            }


            for(i = 0; i < 5; i++) {
                SV1[i] = RndA[i];
                SV1[5 + i] = RndB[i];
                SV1[10 + i] = (byte)(RndA[7 + i] ^ RndB[7 + i]);
                SV2[i] = RndA[4 + i];
                SV2[5 + i] = RndB[4 + i];
                SV2[10 + i] = (byte)(RndA[11 + i] ^ RndB[11 + i]);
            }
            SV1[15] = (byte)0x81;
            SV2[15] = (byte)0x82;

            switch(KeyKind)
            {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    mSessionKeyEnc = Helper.AES.aes192Encrypt24( SV1, Kx );
                    mSessionKeyMAC = Helper.AES.aes192Encrypt24( SV2, Kx );
                    mKeyX = Arrays.copyOf(Kx, 24);
                    break;
                default:
                    mSessionKeyEnc = Helper.AES.aes128Encrypt16( SV1, Kx );
                    mSessionKeyMAC = Helper.AES.aes128Encrypt16( SV2, Kx );
                    mKeyX = Arrays.copyOf(Kx, 16);
            }


            return true;
        }//samAuthenticateHost

        private boolean samLockUnlock(int logicalChannel,
                                      int subCmd,
                                      int keyKind,
                                      int keyNo,
                                      int keyVer,
                                      byte[] kx,
                                      int unlockKeyNo,
                                      int unlockKeyVer,
                                      int maxChainBlock )
        {
            int i, j;
            int sendLen = 0;
            byte[] toSM;
            byte[] from;
            byte[] SV1 = new byte[16];
            byte[] MAC16 = new byte[16];
            byte[] MAC24 = new byte[24];
            byte[] kxe = new byte[24];
            char status = 0;

            if( subCmd > 3 ) {
                mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                return false;
            }

            // Rnd1 p.6 to SAM
            byte[] Rnd1 = new byte[]{0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B};
            (new Random()).nextBytes(Rnd1);

            // RndA p.14 to SAM
            byte[] RndA = new byte[]{0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F};
            (new Random()).nextBytes(RndA);

            toSM = mSAM_Plain;
            from = mSAM_Coded;
            toSM[sendLen++] = (byte)((byte)0x80 | (byte)logicalChannel);
            toSM[sendLen++] = (byte)0x10;
            toSM[sendLen++] = (byte)subCmd; // P1
            toSM[sendLen++] = 0x00; // P2
            toSM[sendLen++] = (byte)(0x02 + ((subCmd == 0x02) ? 0x02 : ((subCmd == 0x03) ? 0x03 : 0x00))); // Lc
            toSM[sendLen++] = (byte)keyNo;
            toSM[sendLen++] = (byte)keyVer;
            if( subCmd == 0x02 ) {
                toSM[sendLen++] = (byte)unlockKeyNo;
                toSM[sendLen++] = (byte)unlockKeyVer;
            } else if( subCmd == 0x03 ) {
                toSM[sendLen++] = (byte)(maxChainBlock & 0xFF);
                toSM[sendLen++] = (byte)((maxChainBlock  >> 8) & 0xFF);
                toSM[sendLen++] = (byte)((maxChainBlock  >> 16) & 0xFF);
            }
            toSM[sendLen++] = 0x00;    // Le

            from = transceiveAPDU(Arrays.copyOf(toSM, sendLen));

            if( from == null )
                return false;
            if( ((SW[0] & 0xFF) != 0x90) || ((SW[1] & 0xFF) != 0xAF) ) {
                mLastError = UEM_ERROR_SAM_APDU;
                return false;
            }
            if( from.length == 0 ) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return false;
            }

            for(i = 0; i < 5; i++) {
                SV1[i] = Rnd1[7 + i];
                SV1[5 + i] = from[7 + i];
                SV1[10 + i] = (byte)(from[i] ^ Rnd1[i]);
            }
            SV1[15] = (byte)0x91;

            switch(keyKind) {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    kxe = Helper.AES.aes192Encrypt24( SV1, kx );
                    break;
                default:
                    kxe = Helper.AES.aes128Encrypt16( SV1, kx );
            }

            byte[] MacData = new byte[from.length + 4];
            System.arraycopy(from, 0, MacData, 0, from.length);
            MacData[from.length] = (byte)subCmd;
            if( subCmd == 0x03 ) {
                MacData[from.length + 1] = (byte) (maxChainBlock & 0xFF);
                MacData[from.length + 2] = (byte) ((maxChainBlock >> 8) & 0xFF);
                MacData[from.length + 3] = (byte) ((maxChainBlock >> 16) & 0xFF);
            } else if( subCmd == 0x02 ) {
                MacData[from.length + 1] = (byte)unlockKeyNo;
                MacData[from.length + 2] = (byte)unlockKeyVer;
                MacData[from.length + 3] = 0;
            } else {
                MacData[from.length + 1] = 0;
                MacData[from.length + 2] = 0;
                MacData[from.length + 3] = 0;
            }

            switch(keyKind) {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    MAC24 = Helper.AES.aes192CbcEnMac( MacData, 24, kx, MAC24 );
                    break;
                default:
                    MAC16 = Helper.AES.aes128CbcEnMAC( MacData, 16, kx, MAC16 );
            }

            // p.7
            sendLen = 0;
            toSM = mSAM_Plain;
            toSM[sendLen++] = (byte)(0x80 | logicalChannel);
            toSM[sendLen++] = (byte)0x10;
            toSM[sendLen++] = 0x00; // P1
            toSM[sendLen++] = 0x00; // P2
            toSM[sendLen++] = 0x14; // Lc
            switch(keyKind) {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    for( i = 0, j = 1; i < 12; i++, j+=2 )
                        toSM[sendLen++] = MAC24[j];
                    break;
                default:
                    for( i = 0, j = 1; i < 8; i++, j+=2 )
                        toSM[sendLen++] = MAC16[j];
            }
            System.arraycopy(Rnd1, 0, toSM, sendLen, Rnd1.length);
            sendLen += Rnd1.length;
            toSM[sendLen++] = 0x00;    // Le

            from = transceiveAPDU(Arrays.copyOf(toSM, sendLen));

            if( from == null )
                return false;
            if( ((SW[0] & 0xFF) != 0x90) || ((SW[1] & 0xFF) != 0xAF) ) {
                mLastError = UEM_ERROR_SAM_APDU;
                return false;
            }
            if( from.length == 0 ) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return false;
            }

            byte[] TmpData = new byte[16];
            System.arraycopy(Rnd1, 0, TmpData, 0, Rnd1.length);
            TmpData[Rnd1.length] = (byte)subCmd;
            if( subCmd == 0x03 ) {
                TmpData[Rnd1.length + 1] = (byte) (maxChainBlock & 0xFF);
                TmpData[Rnd1.length + 2] = (byte) ((maxChainBlock >> 8) & 0xFF);
                TmpData[Rnd1.length + 3] = (byte) ((maxChainBlock >> 16) & 0xFF);
            } else if( subCmd == 0x02 ) {
                TmpData[Rnd1.length + 1] = (byte)unlockKeyNo;
                TmpData[Rnd1.length + 2] = (byte)unlockKeyVer;
                TmpData[Rnd1.length + 3] = 0;
            } else {
                TmpData[Rnd1.length + 1] = 0;
                TmpData[Rnd1.length + 2] = 0;
                TmpData[Rnd1.length + 3] = 0;
            }

            byte[] TestMAC = new byte[8];
            switch(keyKind) {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    MAC24 = Helper.AES.aes192CbcEnMac( TmpData, 24, kx, MAC24 );

                    for( i = 0, j = 1; i < 8; i++, j+=3 )
                        TestMAC[i] = MAC24[j];

                    break;
                default:
                    MAC16 = Helper.AES.aes128CbcEnMAC( TmpData, 16, kx, MAC16 );

                    for( i = 0, j = 1; i < 8; i++, j+=2 )
                        TestMAC[i] = MAC16[j];
            }

            if( !Arrays.equals(TestMAC, Arrays.copyOf(from, TestMAC.length)) ) {
                mLastError = UEM_ERROR_SAM_INVALID_MAC;
                return false;
            }

            byte[] RndB;
            switch(keyKind) {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    RndB = Helper.AES.aes192Decrypt24( Arrays.copyOfRange(from, 16, from.length), kxe );
                    break;
                default:
                    RndB = Helper.AES.aes128Decrypt16( Arrays.copyOfRange(from, 8, from.length), kxe );
            }

            byte[] RndARndBB = new byte[RndA.length + RndB.length];
            System.arraycopy(RndA, 0, RndARndBB, 0, RndA.length);
            System.arraycopy(RndB, 2, RndARndBB, RndA.length, RndB.length - 2);
            System.arraycopy(RndB, 0, RndARndBB, RndA.length + RndB.length - 2, 2);

            sendLen = 0;
            toSM = mSAM_Plain;
            toSM[sendLen++] = (byte)(0x80 | logicalChannel);
            toSM[sendLen++] = (byte)0x10;
            toSM[sendLen++] = 0x00; // P1
            toSM[sendLen++] = 0x00; // P2
            toSM[sendLen++] = 0x20; // Lc
            byte[] EncRndARndBB = null;
            switch(keyKind)
            {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    byte[] IV24 = new byte[24];
                    Arrays.fill(IV24, (byte)0x00);
                    EncRndARndBB = Helper.AES.aes192EncryptCbc(
                            RndARndBB,
                            RndARndBB.length,
                            kxe,
                            IV24,
                            null );
                    break;
                default:
                    byte[] IV16 = new byte[16];
                    Arrays.fill(IV16, (byte)0x00);
                    EncRndARndBB = Helper.AES.aes128EncryptCbc(
                            RndARndBB,
                            RndARndBB.length,
                            kxe,
                            IV16,
                            null );
            }
            System.arraycopy(EncRndARndBB, 0, toSM, sendLen, EncRndARndBB.length);
            sendLen += EncRndARndBB.length;
            toSM[sendLen++] = 0x00;    // Le

            from = transceiveAPDU(Arrays.copyOf(toSM, sendLen));

            if( from == null )
                return false;
            if( ((SW[0] & 0xFF) != 0x90) || ((SW[1] & 0xFF) != 0x00) ) {
                mLastError = UEM_ERROR_SAM_APDU;
                return false;
            }
            if( from.length == 0 ) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return false;
            }

            //if ( status == 0x01)
            //    return S_OK;

            byte[] TestRndAA = null;
            switch(keyKind)
            {
                case UEM_SAM_KEY_KIND_TDEA_ISO10116_3KEY:
                case UEM_SAM_KEY_KIND_AES192:
                    TestRndAA = Helper.AES.aes192Decrypt24( from, kxe );
                    break;
                default:
                    TestRndAA = Helper.AES.aes128Decrypt16( from, kxe );
            }

            byte[] RndAA = new byte[RndA.length];
            System.arraycopy(RndA, 2, RndAA, 0, RndAA.length - 2);
            System.arraycopy(RndA, 0, RndAA, RndAA.length - 2, 2);

            if( !Arrays.equals(RndAA, TestRndAA) ) {
                mLastError = UEM_ERROR_SAM_AUTHENTICATION_FAILED;
                return false;
            }

            return true;
        }
    }

    public CommandsSAM SAM = new CommandsSAM();

    public class CommandsCards {
        public static final int UEM_CARD_STANDARD_ISO14443A = 0x00;
        public static final int UEM_CARD_STANDARD_ISO14443B = 0x10;
        public static final int UEM_CARD_STANDARD_ISO15693 = 0x40;

        public static final int UEM_ISO14443_BAUDRATE_106 = (byte)0x00;
        public static final int UEM_ISO14443_BAUDRATE_212 = (byte)0x01;
        public static final int UEM_ISO14443_BAUDRATE_424 = (byte)0x02;
        public static final int UEM_ISO14443_BAUDRATE_848 = (byte)0x03;

        public static final int UEM_ISO14443TCL_EXCHANGE_DEFAULT = (byte)0x00;
        public static final int UEM_ISO14443TCL_EXCHANGE_TXCHAINING = (byte)0x01;
        public static final int UEM_ISO14443TCL_EXCHANGE_RXCHAINING = (byte)0x02;
        public static final int UEM_ISO14443TCL_EXCHANGE_RXCHAINING_BUFSIZE = (byte)0x03;
        private static final int UEM_ISO14443TCL_EXCHANGE_TX_CRC = 0x0010;
        private static final int UEM_ISO14443TCL_EXCHANGE_RX_CRC = 0x0020;
        private static final int UEM_ISO14443TCL_EXCHANGE_PARITY = 0x0040;
        private static final int UEM_ISO14443TCL_EXCHANGE_LEAVE_BUFFER_BIT = 0x4000;
        private static final int UEM_ISO14443TCL_EXCHANGE_BUFFERED_BIT = 0x8000;

        public byte[] directIOCard(byte[] command) {
            return directIOCard(command, 1000);
        }

        public byte[] directIOCard(byte[] command, int timeout) {
            if (command == null || command.length == 0) {
                mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                return null;
            }
            ByteArrayOutputStream commandReader = new ByteArrayOutputStream();
            commandReader.write(0x48);
            commandReader.write((byte)(command.length & 0xFF));
            commandReader.write((byte)((command.length & 0xFFFF) >> 8));
            commandReader.write(command, 0, command.length);
            commandReader.write((byte)(timeout & 0xFF));
            commandReader.write((byte)((timeout & 0xFFFF) >> 8));
            commandReader.write((byte)((timeout & 0xFFFFFF) >> 16));
            commandReader.write((byte)(timeout >> 24));
            byte[] res = dispatchCommand(commandReader.toByteArray());
            if (!getLastError().equals(UEM_STATUS_SUCCESS))
                return null;
            return Arrays.copyOfRange(res, 2, res.length);
        }

        public Object activateExISO14443(int cardStandard, int cardID, int transmitBaudrate, int receiveBaudrate, int rfOffTime, int rfOnPause, boolean disableTCLSwitch, byte AFI, boolean extATQ, byte SNI) {
            if (cardStandard != UEM_CARD_STANDARD_ISO14443A && cardStandard != UEM_CARD_STANDARD_ISO14443B) {
                mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                return null;
            }

            byte baud = (byte)(transmitBaudrate & 0x03 | ((receiveBaudrate & 0x03) << 2));

            ByteArrayOutputStream commandReader = new ByteArrayOutputStream();
            commandReader.write(0x75);
            commandReader.write((byte)(((cardStandard == UEM_CARD_STANDARD_ISO14443B) ? 0x10 : 0x00) | (baud & 0x0F)));
            commandReader.write((byte)((rfOffTime << 4) | (rfOnPause & 0x0F)));
            commandReader.write((byte)((disableTCLSwitch ? 0x80 : 0x00) | ((byte)cardID & 0x0F)));
            if (cardStandard == UEM_CARD_STANDARD_ISO14443B) {
                commandReader.write((byte)AFI);
                commandReader.write((byte)((extATQ ? 0x10 : 0x00) | (SNI & 0x07)));
            }
            byte[] res = dispatchCommand(commandReader.toByteArray());
            if (res == null || !getLastError().equals(UEM_STATUS_SUCCESS)) {
                return null;
            }

            Object tag = null;
            if (cardStandard == UEM_CARD_STANDARD_ISO14443A) {
                tag = this.ISO14443A.new TagISO14443A();

                ((CommandsISO14443A.TagISO14443A)tag).ATQ = Arrays.copyOfRange(res, 0, 2);
                ((CommandsISO14443A.TagISO14443A)tag).SAK = res[2];
                int UIDLength = res[3];
                ((CommandsISO14443A.TagISO14443A)tag).UID = Arrays.copyOfRange(res, 4, 4 + UIDLength);

                if (!disableTCLSwitch) {
                    int ATSSize = res[4 + UIDLength];
                    ((CommandsISO14443A.TagISO14443A)tag).ATS = Arrays.copyOfRange(res, 4 + UIDLength, 4 + UIDLength + ATSSize);
                }
                this.ISO14443A.mLastUID = Arrays.copyOf(((CommandsISO14443A.TagISO14443A)tag).UID, ((CommandsISO14443A.TagISO14443A)tag).UID.length);
                this.ISO14443A.mLastATQ = Arrays.copyOf(((CommandsISO14443A.TagISO14443A)tag).ATQ, ((CommandsISO14443A.TagISO14443A)tag).ATQ.length);
                this.ISO14443A.mLastSAK = ((CommandsISO14443A.TagISO14443A)tag).SAK;
            } else {
                tag = this.ISO14443B.new TagISO14443B();

                ((CommandsISO14443B.TagISO14443B)tag).MBLI = res[0];
                int ATQLength = res[1];
                ((CommandsISO14443B.TagISO14443B)tag).ATQ = Arrays.copyOfRange(res, 2, 2 + ATQLength);
                if (ATQLength > 8) {
                    ((CommandsISO14443B.TagISO14443B)tag).PUPI = Arrays.copyOfRange(((CommandsISO14443B.TagISO14443B)tag).ATQ, 0, 4);
                    ((CommandsISO14443B.TagISO14443B)tag).appData = Arrays.copyOfRange(((CommandsISO14443B.TagISO14443B)tag).ATQ, 4, 4 + 4);
                    ((CommandsISO14443B.TagISO14443B)tag).protInfo = Arrays.copyOfRange(((CommandsISO14443B.TagISO14443B)tag).ATQ, 4 + 4, ATQLength);
                }
            }

            return tag;
        }

        public byte[] exchangeISO14443TCL(byte[] command) {
            return exchangeISO14443TCL(command, UEM_ISO14443TCL_EXCHANGE_DEFAULT, false, false, false, false, false);
        }

        public byte[] exchangeISO14443TCL(byte[] command, int mode, boolean txCRC, boolean rxCRC, boolean parity, boolean leaveBufferBit, boolean exchangeBufferedBit) {
            if (command == null || command.length == 0) {
                mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                return null;
            }
            int option = (int)mode +
                    (txCRC?UEM_ISO14443TCL_EXCHANGE_TX_CRC:0) +
                    (rxCRC?UEM_ISO14443TCL_EXCHANGE_RX_CRC:0) +
                    (parity?UEM_ISO14443TCL_EXCHANGE_PARITY:0) +
                    (leaveBufferBit?UEM_ISO14443TCL_EXCHANGE_LEAVE_BUFFER_BIT:0) +
                    (exchangeBufferedBit?UEM_ISO14443TCL_EXCHANGE_BUFFERED_BIT:0);
            ByteArrayOutputStream commandReader = new ByteArrayOutputStream();
            commandReader.write(0x47);
            commandReader.write((byte)(option & 0xFF));
            commandReader.write((byte)((option >> 8) & 0xFF));
            commandReader.write((byte)(command.length & 0xFF));
            commandReader.write((byte)((command.length & 0xFFFF) >> 8));
            commandReader.write(command, 0, command.length);
            byte[] res = dispatchCommand(commandReader.toByteArray());
            if (!getLastError().equals(UEM_STATUS_SUCCESS))
                return null;
            if (res.length < 2) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return null;
            }

            int len = res[0] + res[1] * 256;
            if (res.length != 2 + len) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return null;
            }
            return Arrays.copyOfRange(res, 2, res.length);
        }

        public class CommandsISO14443A {
            public byte[] mLastATQ = new byte[2];
            public byte mLastSAK = 0x00;
            public byte[] mLastUID = new byte[10];

            public class TagISO14443A {
                public byte[] ATQ = new byte[2];
                public byte SAK = 0x00;
                public byte[] UID = new byte[10];
                public byte[] ATS = null;
            }

            public TagISO14443A activateIdle() {
                byte[] res = dispatchCommand(new byte[]{0x43});
                if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                    //mLastError = UEM_ERROR_READER_RESPONSE_FAILURE;
                    return null;
                }

                if (res.length < 4) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                if (res.length < 4 + res[4]) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                TagISO14443A tag = new TagISO14443A();

                mLastATQ[0] = res[0];
                tag.ATQ[0] = res[0];
                mLastATQ[1] = res[1];
                tag.ATQ[1] = res[1];
                mLastSAK = res[2];
                tag.SAK = res[2];
                byte uidLength = res[3];

                mLastUID = Arrays.copyOfRange(res, 4, res.length);
                tag.UID = Arrays.copyOfRange(res, 4, res.length);

                return tag;
            }

            public boolean halt() {
                dispatchCommand(new byte[]{0x1d});
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public boolean wakeUp() {
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0x44);
                command.write((byte)(mLastUID.length & 0xFF));
                command.write(mLastUID, 0, mLastUID.length);
                byte[] res = dispatchCommand(command.toByteArray());
                if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                    return false;
                }

                if (res.length < 4) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return false;
                }

                mLastATQ[0] = res[0];
                mLastATQ[1] = res[1];
                mLastSAK = res[2];

                return true;
            }

            /*public byte[] RATS() {
                return RATS((byte)0);
            }

            public byte[] RATS(byte cardID) {
                byte[] res = dispatchCommand(new byte[]{0x35, cardID});
                if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                    return null;
                }

                if (res.length < 2) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                int len = res[0] + res[1] * 256;

                if (len == 0 || res.length - 2 < len){
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                return Arrays.copyOfRange(res, 2, res.length);
            }

            public boolean PPS() {
                return PPS((byte)0, UEM_ISO14443_BAUDRATE_106, UEM_ISO14443_BAUDRATE_106);
            }

            public boolean PPS(byte cardID, byte transmitBaudrate, byte receiveBaudrate) {
                byte baud = (byte)(transmitBaudrate & 0x03 | ((receiveBaudrate & 0x03) << 2));
                dispatchCommand(new byte[] {0x36, cardID, baud});
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }*/

            public TagISO14443A activateEx() {
                return activateEx(true);
            }

            public TagISO14443A activateExTCL() {
                return activateEx(false);
            }

            public TagISO14443A activateEx(boolean disableTCLSwitch) {
                return activateEx(UEM_ISO14443_BAUDRATE_106, UEM_ISO14443_BAUDRATE_106, 10, 10, disableTCLSwitch);
            }

            public TagISO14443A activateEx(int transmitBaudrate, int receiveBaudrate, int rfOffTime, int rfOnPause, boolean disableTCLSwitch) {
                return activateEx(0, transmitBaudrate, receiveBaudrate, rfOffTime, rfOnPause, disableTCLSwitch);
            }

            public TagISO14443A activateEx(int cardID, int transmitBaudrate, int receiveBaudrate, int rfOffTime, int rfOnPause, boolean disableTCLSwitch) {
                return (TagISO14443A)activateExISO14443(
                        UEM_CARD_STANDARD_ISO14443A,
                        cardID,
                        transmitBaudrate,
                        receiveBaudrate,
                        rfOffTime,
                        rfOnPause,
                        disableTCLSwitch,
                        (byte)0x00,
                        false,
                        (byte)0x00);
            }

            public byte[] activateTCLdeprecated() {
                return activateTCLdeprecated(0, UEM_ISO14443_BAUDRATE_106, UEM_ISO14443_BAUDRATE_106);
            }

            public byte[] activateTCLdeprecated(int cardID, int transmitBaudrate, int receiveBaudrate) {
                byte baud = (byte)(transmitBaudrate & 0x03 | ((receiveBaudrate & 0x03) << 2));
                byte[] res = dispatchCommand(new byte[]{0x2B, (byte)cardID, baud});
                if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                    return null;
                }

                if (res.length < 2) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                int len = res[0] + res[1] * 256;

                if (len == 0 || res.length - 2 != len){
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                return Arrays.copyOfRange(res, 2, res.length);
            }
        }

        public CommandsISO14443A ISO14443A = new CommandsISO14443A();

        public class CommandsISO14443B {

            public byte[] mLastPUPI = new byte[4];
            public byte[] mLastAppData = new byte[4];
            public byte[] mLastProtInfo = new byte[3];

            public class TagISO14443B {
                public byte[] PUPI = new byte[4];
                public byte[] appData = new byte[4];
                public byte[] protInfo = new byte[3];
                public byte MBLI = 0x00;
                public byte[] ATQ = null;
            }

            public class AttribISO14443B {
                public byte[] higherLayerBuffer = new byte[0];
                public byte mbliCid;
            }

            public TagISO14443B[] activateIdle() {
                return activateIdle((byte)0x00);
            }

            public TagISO14443B[] activateIdle(byte afi) {
                return activateIdle(afi, (byte)0x02);
            }

            public TagISO14443B[] activateIdle(byte afi, byte sni) {
                return activateIdle(afi, sni, false);
            }

            private TagISO14443B[] activateIdle(byte afi, byte sni, boolean isWakeUp) {
                ArrayList<TagISO14443B> tags = new ArrayList<TagISO14443B>();

                byte[] res = dispatchCommand(new byte[]{isWakeUp?(byte)0x57:(byte)0x56, afi, sni});
                if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                    return null;
                }

                if (res.length < 11) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                int len = res[0] + res[1] * 256;
                if (res.length % 11 != 0) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                int i = 0;
                while (i < res.length) {
                    TagISO14443B tag = new TagISO14443B();
                    System.arraycopy(res, i, tag.PUPI, 0, 4);
                    System.arraycopy(res, i, mLastPUPI, 0, 4);
                    i += 4;
                    System.arraycopy(res, i, tag.appData, 0, 4);
                    System.arraycopy(res, i, mLastAppData, 0, 4);
                    i += 4;
                    System.arraycopy(res, i, tag.protInfo, 0, 3);
                    System.arraycopy(res, i, mLastProtInfo, 0, 3);
                    i += 3;
                    tags.add(tag);
                }
                return tags.toArray(new CommandsISO14443B.TagISO14443B[0]);
            }

            public boolean halt() {
                return halt(mLastPUPI);
            }

            public boolean halt(byte[] PUPI) {
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0x55);
                command.write(PUPI, 0, PUPI.length);
                dispatchCommand(command.toByteArray());
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public TagISO14443B[] wakeUp() {
                return wakeUp((byte)0x00);
            }

            public TagISO14443B[] wakeUp(byte afi) {
                return wakeUp(afi, (byte)0x02);
            }

            public TagISO14443B[] wakeUp(byte afi, byte sni) {
                return activateIdle(afi, sni, true);
            }

            public AttribISO14443B attribDeprecated(byte param1, byte param2, byte param3, byte param4, byte[] higherLayerBuffer) {
                return attribDeprecated(mLastPUPI, param1, param2, param3, param4, higherLayerBuffer);
            }

            public AttribISO14443B attribDeprecated(byte[] PUPI, byte param1, byte param2, byte param3, byte param4, byte[] higherLayerBuffer) {
                int bufLen = (higherLayerBuffer != null) ? higherLayerBuffer.length : 0;

                if (PUPI == null || PUPI.length != 4 ) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return null;
                }

                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0x54);
                command.write(8 + bufLen);
                command.write(PUPI, 0, PUPI.length);
                command.write(param1);
                command.write(param2);
                command.write(param3);
                command.write(param4);
                if (bufLen > 0)
                    command.write(higherLayerBuffer, 0, higherLayerBuffer.length);
                byte[] res = dispatchCommand(command.toByteArray());
                if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                    return null;
                }
                if (res == null || (res.length < 1)) {
                    mLastError = UEM_ERROR_COMMUNICATION_DATA_LOST;
                    return null;
                }
                AttribISO14443B attrib = new AttribISO14443B();
                attrib.mbliCid = res[0];

                if ((attrib.mbliCid & 0x0f) != param4) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;//Status = SCARD_F_INTERNAL_ERROR;
                    return null;
                }

                if (res.length > 1) {
                    attrib.higherLayerBuffer = Arrays.copyOfRange(res, 1, res.length);
                }

                return attrib;
            }

            public TagISO14443B activateEx() {
                return activateEx(true);
            }

            public TagISO14443B activateExTCL() {
                return activateEx(false);
            }

            public TagISO14443B activateEx(boolean disableTCLSwitch) {
                return activateEx(UEM_ISO14443_BAUDRATE_106, UEM_ISO14443_BAUDRATE_106, 10, 10, disableTCLSwitch);
            }

            public TagISO14443B activateEx(int transmitBaudrate, int receiveBaudrate, int rfOffTime, int rfOnPause, boolean disableTCLSwitch) {
                return activateEx(transmitBaudrate, receiveBaudrate, rfOffTime, rfOnPause, disableTCLSwitch, (byte)0x00, false, (byte)0x02);
            }

            public TagISO14443B activateEx(int transmitBaudrate, int receiveBaudrate, int rfOffTime, int rfOnPause, boolean disableTCLSwitch, byte AFI, boolean extATQ, byte SNI) {
                return activateEx(0, transmitBaudrate, receiveBaudrate, rfOffTime, rfOnPause, disableTCLSwitch, AFI, extATQ, SNI);
            }

            public TagISO14443B activateEx(int cardID, int transmitBaudrate, int receiveBaudrate, int rfOffTime, int rfOnPause, boolean disableTCLSwitch, byte AFI, boolean extATQ, byte SNI) {
                return (TagISO14443B)activateExISO14443(
                        UEM_CARD_STANDARD_ISO14443B,
                        cardID,
                        transmitBaudrate,
                        receiveBaudrate,
                        rfOffTime,
                        rfOnPause,
                        disableTCLSwitch,
                        AFI,
                        extATQ,
                        SNI);
            }
        }

        public CommandsISO14443B ISO14443B = new CommandsISO14443B();

        public class CommandsISO15693 {

            public class TagISO15693 {
                public int retCode;
                public int size;
                public int flags;
                public int DSFID;
                public byte[] UID;
                //public byte[] crc16;
            }

            public class BlockISO15693 {
                public int flags;
                public int securityStatus;
                public byte[] data;
                public String error;
            }

            public byte[] mUID = new byte[8];

            public TagISO15693[] findAllTags() {
                return findAllTags((byte)0x00);
            }

            public TagISO15693[] findAllTags(int AFI) {
                byte[] res = dispatchCommand(new byte[]{0x30, 0x00, 0x00, (byte)AFI});
                if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                    return null;
                }

                if (res.length < 2) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                int len = res[0] + res[1] * 256;
                if (res.length != 2 + len) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                ArrayList<TagISO15693> tags = new ArrayList<TagISO15693>();

                int i = 0;
                while (i < res.length - 2) {
                    TagISO15693 tag = new TagISO15693();
                    tag.retCode = res[2+i++];
                    if (tag.retCode == 0x00) {
                        tag.size = res[2+i++];
                        if (tag.size >= 2) {
                            tag.flags = res[2+i++];
                            tag.DSFID = res[2+i++];
                            if (tag.size > 2) {
                                tag.UID = new byte[tag.size - 2];
                                System.arraycopy(res, 2 + i, tag.UID, 0, tag.size - 2);
                                i += tag.size - 2;
                                System.arraycopy(tag.UID, 0, mUID, 0, tag.UID.length);
                            }
                        }
                    } else {
//                        tag.crc16 = new byte[2];
//                        System.arraycopy(res, 2+i, tag.crc16, 0, 2);
//                        i += 2;
                    }
                    tags.add(tag);
                }
                return tags.toArray(new TagISO15693[0]);
            }

            private String translateErrorISO15693(int errorCode)
            {
                String Error;

                switch (errorCode)
                {
                    case 0x01:
                        Error = UEM_ERROR_UNSUPPORTED_FEATURE;//SCARD_E_UNSUPPORTED_FEATURE;
                        break;
                    case 0x02:
                        Error = UEM_ERROR_COMMUNICATION_DATA_LOST;//SCARD_E_COMM_DATA_LOST;
                        break;
                    case 0x03:
                        Error = UEM_ERROR_INCORRECT_PARAMETER;//SCARD_E_INVALID_PARAMETER;
                        break;
                    case 0x0F:
                        Error = UEM_ERROR_UNEXPECTED;//SCARD_E_UNEXPECTED;
                        break;
                    case 0x010:
                        Error = UEM_ERROR_INCORRECT_PARAMETER;//SCARD_E_INVALID_PARAMETER;
                        break;
                    case 0x011:
                        Error = UEM_ERROR_ACCESS;//SCARD_E_NO_ACCESS;
                        break;
                    case 0x012:
                        Error = UEM_ERROR_ACCESS;//SCARD_E_NO_ACCESS;
                        break;
                    case 0x013:
                        Error = UEM_ERROR_NOT_TRANSACTED;//SCARD_E_NOT_TRANSACTED;
                        break;
                    case 0x014:
                        Error = UEM_ERROR_NOT_TRANSACTED;//SCARD_E_NOT_TRANSACTED;
                        break;
                    default:
                        Error = UEM_ERROR_UNEXPECTED;//SCARD_E_UNEXPECTED;
                }
                if (((errorCode & 0xFF) > 0xA0) && ((errorCode & 0xFF) < 0xDF))
                    Error = UEM_ERROR_UNEXPECTED;//SCARD_E_UNEXPECTED;

                return Error;
            }

            public BlockISO15693 select() {
                return select(Cards.ISO15693.mUID);
            }

            public BlockISO15693 select(byte[] UID) {
                if (UID == null || UID.length == 0) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return null;
                }

                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0x22);
                command.write(0x25);
                command.write(UID, 0, UID.length);

                byte[] res = Cards.directIOCard(command.toByteArray(), 2500);

                if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                    return null;
                }
                if (res.length < 1) {
                    mLastError = UEM_ERROR_COMMUNICATION_DATA_LOST;
                    return null;
                }

                BlockISO15693 block = new BlockISO15693();

                block.flags = res[0];
                if ((block.flags & 0x01) == 0x01) {
                    if (res.length < 2) {
                        mLastError = UEM_ERROR_COMMUNICATION_DATA_LOST;
                        return null;
                    }
                    block.error = translateErrorISO15693(res[1]);
                }

                return block;
            }

            public BlockISO15693 readSingleBlock(int blockNumber) {
                return readSingleBlock(blockNumber, false);
            }

            public BlockISO15693 readSingleBlock(int blockNumber, boolean optionSecurity) {
                return readSingleBlock(blockNumber, optionSecurity, true, Cards.ISO15693.mUID);
            }

            public BlockISO15693 readSingleBlock(int blockNumber, boolean optionSecurity, boolean addressMode, byte[] UID) {
                //boolean addressMode = (flags & 0x20) == 0x20;
                //boolean optionSecurity = (flags & 0x40) == 0x40;

                if (addressMode && (UID == null)) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return null;
                }

                byte flags = (byte)(0x02 + (addressMode?0x20:0x10) + (optionSecurity?0x40:0x00));

                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(flags);
                command.write(0x20);
                if (addressMode)
                    command.write(UID, 0, UID.length);
                command.write(blockNumber);

                byte[] res = Cards.directIOCard(command.toByteArray(), 1200);

                if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                    return null;
                }
                if (res.length < 2) {
                    mLastError = UEM_ERROR_COMMUNICATION_DATA_LOST;
                    return null;
                }

                BlockISO15693 block = new BlockISO15693();

                block.flags = res[0];

                if ((block.flags & 0x01) == 0x01) {
			        block.error = translateErrorISO15693(res[1]);
                } else {
                    int expected = optionSecurity ? 6 : 5;
                    if (res.length < expected) {
                        mLastError = UEM_ERROR_COMMUNICATION_DATA_LOST;//Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
                        return null;
                    }

                    int index = 1;
                    if (optionSecurity)
					    block.securityStatus = res[index++];
                    block.data = new byte[4];
                    System.arraycopy(res, index, block.data, 0, block.data.length);
                }

                return block;
            }

            public BlockISO15693 writeSingleBlock(int blockNumber, byte[] data) {
                return writeSingleBlock(blockNumber, data, false);
            }

            public BlockISO15693 writeSingleBlock(int blockNumber, byte[] data, boolean optionSecurity) {
                return writeSingleBlock(blockNumber, data, optionSecurity, true, Cards.ISO15693.mUID);
            }

            public BlockISO15693 writeSingleBlock(int blockNumber, byte[] data, boolean optionSecurity, boolean addressMode, byte[] UID) {
                //boolean addressMode = (flags & 0x20) == 0x20;

                if (addressMode && (UID == null)) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return null;
                }

                byte flags = (byte)(0x02 + (addressMode?0x20:0x10) + (optionSecurity?0x40:0x00));

                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(flags);
                command.write(0x21);
                if (addressMode)
                    command.write(UID, 0, UID.length);
                command.write(blockNumber);
                command.write(data, 0, data.length);

                byte[] res = Cards.directIOCard(command.toByteArray(), 2500);

                if (!getLastError().equals(UEM_STATUS_SUCCESS)) {
                    return null;
                }
                if (res.length < 1) {
                    mLastError = UEM_ERROR_COMMUNICATION_DATA_LOST;
                    return null;
                }

                BlockISO15693 block = new BlockISO15693();

                block.flags = res[0];
                if ((block.flags & 0x01) == 0x01) {
                    if (res.length < 2) {
                        mLastError = UEM_ERROR_COMMUNICATION_DATA_LOST;
                        return null;
                    }
                    block.error = translateErrorISO15693(res[1]);
                }

                return block;
            }
        }

        public CommandsISO15693 ISO15693 = new CommandsISO15693();


        public class CommandsMifareClassic {
            public static final int UEM_MIFARE_KEY_TYPE_A = (byte)0x0A;
            public static final int UEM_MIFARE_KEY_TYPE_B = (byte)0x0B;

            public static final int UEM_MIFARE_AUTHENTICATE_TYPE_FIRST = (byte)0x01;
            public static final int UEM_MIFARE_AUTHENTICATE_TYPE_FOLLOWING = (byte)0x0F;

            public boolean authenticateWithCodedKey(int sector, int keyType, byte[] codedKey) {
                return authenticateWithCodedKey(Cards.ISO14443A.mLastUID, sector, keyType, codedKey);
            }

            public boolean authenticateWithCodedKey(byte[] UID, int sector, int keyType, byte[] codedKey) {
                return authenticateWithCodedKey(UID, (int)Helper.getActualBlockAddress(sector, 0), keyType, codedKey);
            }

            public boolean authenticateWithCodedKey(long absoluteBlockAddress, int keyType, byte[] codedKey) {
                return authenticateWithCodedKey(Cards.ISO14443A.mLastUID, absoluteBlockAddress, keyType, codedKey);
            }

            public boolean authenticateWithCodedKey(byte[] UID, long absoluteBlockAddress, int keyType, byte[] codedKey) {
                if (UID == null || UID.length < 4 || codedKey == null || codedKey.length != 12) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0x18);
                command.write((keyType == 0x0A)?0x60:0x61);
                command.write(UID, UID.length - 4, UID.length);
                command.write(codedKey, 0, codedKey.length);
                command.write((byte)absoluteBlockAddress);
                dispatchCommand(command.toByteArray());
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public boolean authenticateWithPlainKey(int sector, int keyType, byte[] plainKey) {
                return authenticateWithPlainKey(Cards.ISO14443A.mLastUID, sector, keyType, plainKey);
            }

            public boolean authenticateWithPlainKey(byte[] UID, int sector, int keyType, byte[] plainKey) {
                return authenticateWithPlainKey(UID, (int)Helper.getActualBlockAddress(sector, 0), keyType, plainKey);
            }

            public boolean authenticateWithPlainKey(long absoluteBlockAddress, int keyType, byte[] plainKey) {
                return authenticateWithPlainKey(Cards.ISO14443A.mLastUID, absoluteBlockAddress, keyType, plainKey);
            }

            public boolean authenticateWithPlainKey(byte[] UID, long absoluteBlockAddress, int keyType, byte[] plainKey) {
                if (UID == null || UID.length < 4 || plainKey == null || plainKey.length != 6) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0x14);
                command.write((keyType == 0x0A)?0x60:0x61);
                command.write(UID, UID.length - 4, 4);
                command.write(plainKey, 0, plainKey.length);
                command.write((byte)absoluteBlockAddress);
                dispatchCommand(command.toByteArray());
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public boolean writeKeyToEEPROM(int eepromRecord, int keyType, byte[] plainKey) {
                if (plainKey == null || plainKey.length != 6) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0x17);
                command.write((keyType == 0x0A)?0x60:0x61);
                command.write(eepromRecord);
                command.write(plainKey, 0, plainKey.length);
                dispatchCommand(command.toByteArray());
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            /*public boolean authenticateWithEEPROMKey(byte sector, byte keyType) {
                return authenticateWithEEPROMKey(Cards.ISO14443A.mLastUID, sector, keyType);
            }

            public boolean authenticateWithEEPROMKey(byte[] mLastUID, byte sector, byte keyType) {
                return authenticateWithEEPROMKey(mLastUID, sector, sector, keyType);
            }*/

            public boolean authenticateWithEEPROMKey(int eepromRecord, int sector, int keyType) {
                return authenticateWithEEPROMKey(Cards.ISO14443A.mLastUID, eepromRecord, sector, keyType);
            }

            public boolean authenticateWithEEPROMKey(byte[] UID, int eepromRecord, int sector, int keyType) {
                return authenticateWithEEPROMKey(UID, eepromRecord, (int)Helper.getActualBlockAddress(sector, 0), keyType);
            }

            public boolean authenticateWithEEPROMKey(int eepromRecord, long absoluteBlockAddress, int keyType) {
                return authenticateWithEEPROMKey(Cards.ISO14443A.mLastUID, eepromRecord, absoluteBlockAddress, keyType);
            }

            public boolean authenticateWithEEPROMKey(byte[] UID, int eepromRecord, long absoluteBlockAddress, int keyType) {
                if (UID == null || UID.length < 4) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0x15);
                command.write((keyType == 0x0A)?0x60:0x61);
                command.write(UID, UID.length - 4, 4);
                command.write(eepromRecord);
                command.write((byte)absoluteBlockAddress);
                dispatchCommand(command.toByteArray());
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public boolean writeBlock(int sector, int block, byte[] data) {
                return writeBlock(Helper.getActualBlockAddress(sector, block), data);
            }

            public boolean writeBlock(long absoluteBlockAddress, byte[] data) {
                if (data == null || data.length != 16) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0x1A);
                command.write((byte)absoluteBlockAddress);
                command.write(data, 0, data.length);
                dispatchCommand(command.toByteArray());
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public byte[] readBlock(int sector, int block) {
                return readBlock(Helper.getActualBlockAddress(sector, block));
            }

            public byte[] readBlock(long absoluteBlockAddress) {
                return dispatchCommand(new byte[] {0x19, (byte)absoluteBlockAddress});
            }

            public boolean authenticateFirstSAM(int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM(UEM_MIFARE_AUTHENTICATE_TYPE_FIRST, sector, keyType, keyNumber, keyVersion);
            }

            public boolean authenticateFollowingSAM(int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM(UEM_MIFARE_AUTHENTICATE_TYPE_FOLLOWING, sector, keyType, keyNumber, keyVersion);
            }

            public boolean authenticateSAM(int authenticationType, int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM((byte)0x00, authenticationType, sector, keyType, keyNumber, keyVersion);
            }

            public boolean authenticateSAM(int logicalChannel, int authenticationType, int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM(Cards.ISO14443A.mLastUID, logicalChannel, authenticationType, sector, keyType, keyNumber, keyVersion);
            }

            public boolean authenticateSAM(byte[] UID, int logicalChannel, int authenticationType, int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM(UID, logicalChannel, authenticationType, (int)Helper.getActualBlockAddress(sector, 0), keyType, keyNumber, keyVersion);
            }

            public boolean authenticateSAM(int logicalChannel, int authenticationType, long absoluteBlockAddress, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM(Cards.ISO14443A.mLastUID, logicalChannel, authenticationType, absoluteBlockAddress, keyType, keyNumber, keyVersion);
            }

            public boolean authenticateSAM(byte[] UID, int logicalChannel, int authenticationType, long absoluteBlockAddress, int keyType, int keyNumber, int keyVersion) {
                if (UID == null || UID.length < 4) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0xAB);
                command.write(logicalChannel);
                command.write(authenticationType);
                command.write(UID, UID.length - 4, 4);
                command.write(keyNumber);
                command.write(keyVersion);
                command.write(keyType);
                command.write((byte)absoluteBlockAddress);
                byte[] res = dispatchCommand(command.toByteArray());
                if (res != null && res.length >= 2) {
                    SAM.SW[0] = res[0];
                    SAM.SW[1] = res[1];
                }
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public byte[] readBlockSAM(int sector, int blockInSector) {
                return readBlockSAM((byte)0x00, (int)Helper.getActualBlockAddress(sector, blockInSector));
            }

            public byte[] readBlockSAM(int logicalChannel, int sector, int blockInSector) {
                return readBlockSAM(logicalChannel, (int)Helper.getActualBlockAddress(sector, blockInSector));
            }

            public byte[] readBlockSAM(int logicalChannel, long absoluteBlockAddress) {
                byte[] res = dispatchCommand(new byte[]{(byte)0xAC, (byte)logicalChannel, (byte)absoluteBlockAddress});
                if (res != null && res.length >= 2) {
                    SAM.SW[0] = res[0];
                    SAM.SW[1] = res[1];
                    if (getLastError().equals(UEM_STATUS_SUCCESS)) {
                        return Arrays.copyOfRange(res, 2, res.length);
                    }
                }
                return null;
            }

            public boolean writeBlockSAM(int sector, int blockInSector, byte[] data) {
                return writeBlockSAM((byte)0x00, (int)Helper.getActualBlockAddress(sector, blockInSector), data);
            }

            public boolean writeBlockSAM(int logicalChannel, int sector, int blockInSector, byte[] data) {
                return writeBlockSAM(logicalChannel, (int)Helper.getActualBlockAddress(sector, blockInSector), data);
            }

            public boolean writeBlockSAM(int logicalChannel, long absoluteBlockAddress, byte[] data) {
                if (data == null || data.length != 16) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0xAD);
                command.write((byte)logicalChannel);
                command.write((byte)absoluteBlockAddress);
                command.write(data, 0, data.length);
                byte[] res = dispatchCommand(command.toByteArray());
                if (res != null && res.length >= 2) {
                    SAM.SW[0] = res[0];
                    SAM.SW[1] = res[1];
                }
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }
        }

        public CommandsMifareClassic MifareClassic = new CommandsMifareClassic();

        public class CommandsMifareUltralight {
            public boolean writePage(int page, byte[] data) {
                if (data == null || data.length != 4) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0x1E);
                command.write(page);
                command.write(data, 0, data.length);
                dispatchCommand(command.toByteArray());
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public byte[] readPages(int startPage) {
                return dispatchCommand(new byte[] {0x19, (byte)startPage});
            }
        }

        public CommandsMifareUltralight MifareUltralight = new CommandsMifareUltralight();

        public class CommandsMifarePlus {
            /*public static final byte UEM_MIFARE_PLUS_ENCRYPTION_MODE_CMD_DATA = (byte)0x00;
            public static final byte UEM_MIFARE_PLUS_ENCRYPTION_MODE_CMD_DATA_RESP = (byte)0x01;
            public static final byte UEM_MIFARE_PLUS_ENCRYPTION_MODE_CMD = (byte)0x02;
            public static final byte UEM_MIFARE_PLUS_ENCRYPTION_MODE_CMD_RESP = (byte)0x03;
            public static final byte UEM_MIFARE_PLUS_ENCRYPTION_MODE_DATA = (byte)0x04;
            public static final byte UEM_MIFARE_PLUS_ENCRYPTION_MODE_DATA_RESP = (byte)0x05;
            public static final byte UEM_MIFARE_PLUS_ENCRYPTION_MODE_NONE = (byte)0x06;
            public static final byte UEM_MIFARE_PLUS_ENCRYPTION_MODE_RESP = (byte)0x07;*/

            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_DATA_BLOCK = (byte)0x00;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_CONFIGURATION_BLOCK = (byte)0xB0;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_INSTALLATION_IDENTIFIER = (byte)0xB1;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_ATS_INFORMATION = (byte)0xB2;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_FIELD_CONFIGURATION_BLOCK = (byte)0xB3;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_AES_SECTOR_KEY_A = (byte)0x4A;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_AES_SECTOR_KEY_B = (byte)0x4B;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_ORIGINALITY_KEY = (byte)0x80;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_MASTER_KEY = (byte)0x90;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_CONFIGURATION_KEY = (byte)0x91;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_LEVEL_2_SWITCH_KEY = (byte)0x92;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_LEVEL_3_SWITCH_KEY = (byte)0x93;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_SL1_AUTHENTICATION_KEY = (byte)0x94;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_VIRTUAL_CARD_SELECT_KEY = (byte)0xA0;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_PROXIMITY_CHECK_KEY = (byte)0xA1;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_VIRTUAL_CARD_POLLING_ENCRYPTION_KEY = (byte)0xA2;
            public static final int UEM_MIFARE_PLUS_VALUE_TYPE_VIRTUAL_CARD_POLLING_MAC_KEY = (byte)0xA3;

            public static final int UEM_MIFARE_PLUS_AUTHENTICATION_TYPE_RESET = (byte)0x00;
            public static final int UEM_MIFARE_PLUS_AUTHENTICATION_TYPE_FIRST = (byte)0x01;
            public static final int UEM_MIFARE_PLUS_AUTHENTICATION_TYPE_FOLLOWING = (byte)0x0F;

            public static final int UEM_MIFARE_PLUS_SECURITY_LEVEL_1 = (byte)0x00;
            public static final int UEM_MIFARE_PLUS_SECURITY_LEVEL_2 = (byte)0x01;
            public static final int UEM_MIFARE_PLUS_SECURITY_LEVEL_3 = (byte)0x03;

            public boolean writePersoExplicit(int valueType, int sector, int block, byte[] data) {
                if (data == null || data.length != 16) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0xA8);
                command.write(valueType);
                command.write(sector);
                command.write(block);
                command.write(0xFF);
                command.write(0xFF);
                command.write(data, 0, data.length);
                dispatchCommand(command.toByteArray());
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public boolean commitPerso() {
                dispatchCommand(new byte[]{(byte)0xAA, 0x21, 0x43, 0x65, (byte)0x87});
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public boolean authenticateFirst(int sector, int keyType, int keyFlashAddress) {
                return authenticate((byte)0x01, sector, keyType, keyFlashAddress);
            }

            public boolean authenticateFollowing(int sector, int keyType, int keyFlashAddress) {
                return authenticate((byte)0x0F, sector, keyType, keyFlashAddress);
            }

            public boolean authenticate(int authType, int sector, int keyType, int keyFlashAddress) {
                return authenticate(authType, sector, keyType, keyFlashAddress, new byte[]{});
            }

            public boolean authenticate(int authType, int sector, int keyType, int keyFlashAddress, byte[] pcdCap) {
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0xA0);
                command.write(authType);
                command.write(keyType);
                command.write(sector);
                command.write((byte)(keyFlashAddress & 0xFF));
                command.write((byte)((keyFlashAddress >> 8 )& 0xFF));
                if (pcdCap == null)
                    command.write(0x00);
                else {
                    command.write((byte)pcdCap.length);
                    if (pcdCap.length > 0)
                        command.write(pcdCap, 0, pcdCap.length);
                }
                dispatchCommand(command.toByteArray());
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public byte[] readData(int valueType, int sector, int block, int blocksCount) {
                return readData(valueType, sector, block, blocksCount, true, true, true);
            }

            public byte[] readData(int valueType, int sector, int block, int blocksCount, boolean useCommandMAC, boolean useCommandEncryption, boolean useResponseMAC) {
                byte encryptionMode = (byte)((useCommandMAC?0x00:0x04) | (useCommandEncryption?0x00:0x02) | (useResponseMAC?0x01:0x00));
                byte[] res = dispatchCommand(new byte[] {(byte)0xA4, encryptionMode, (byte)valueType, (byte)sector, (byte)block, (byte)blocksCount});
                if (!getLastError().equals(UEM_STATUS_SUCCESS))
                    return null;
                if (res.length < 2) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                int len = res[0] + res[1] * 256;
                if (res.length != 2 + len) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }
                return Arrays.copyOfRange(res, 2, res.length);
            }

            public boolean writeData(int valueType, int sector, int block, int blocksCount, byte[] data) {
                return writeData(valueType, sector, block, blocksCount, data, true, true, true);
            }

            public boolean writeData(int valueType, int sector, int block, int blocksCount, byte[] data, boolean useCommandMAC, boolean useCommandEncryption, boolean useResponseMAC) {
                if (data == null || data.length != 16 * blocksCount) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                byte encryptionMode = (byte)((useCommandMAC?0x00:0x04) | (useCommandEncryption?0x00:0x02) | (useResponseMAC?0x01:0x00));
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0xA5);
                command.write(encryptionMode);
                command.write(valueType);
                command.write(sector);
                command.write(block);
                command.write(blocksCount);
                command.write(data, 0, data.length);
                dispatchCommand(command.toByteArray());
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public boolean authenticateFirstSAM(int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM(UEM_MIFARE_PLUS_AUTHENTICATION_TYPE_FIRST, sector, keyType, keyNumber, keyVersion);
            }

            public boolean authenticateFollowingSAM(int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM(UEM_MIFARE_PLUS_AUTHENTICATION_TYPE_FOLLOWING, sector, keyType, keyNumber, keyVersion);
            }

            public boolean authenticateResetSAM(int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM(UEM_MIFARE_PLUS_AUTHENTICATION_TYPE_RESET, sector, keyType, keyNumber, keyVersion);
            }

            public boolean authenticateSAM(int authenticationType, int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM(authenticationType, UEM_MIFARE_PLUS_SECURITY_LEVEL_3, sector, keyType, keyNumber, keyVersion);
            }

            public boolean authenticateSAM(int authenticationType, int currentSecurityLevel, int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM((byte)0x00, authenticationType, currentSecurityLevel, sector, keyType, keyNumber, keyVersion, null, null);
            }

            public boolean authenticateSAM(int logicalChannel, int authenticationType, int currentSecurityLevel, int sector, int keyType, int keyNumber, int keyVersion) {
                return authenticateSAM(logicalChannel, authenticationType, currentSecurityLevel, sector, keyType, keyNumber, keyVersion, null, null);
            }

            public boolean authenticateSAM(int logicalChannel, int authenticationType, int currentSecurityLevel, int sector, int keyType, int keyNumber, int keyVersion, byte[] diversificationVector, byte[] pcdCapabilities) {
                byte securityLevel = 0x00;
                switch (currentSecurityLevel) {
                    case 0:
                    case 1:
                        securityLevel = UEM_MIFARE_PLUS_SECURITY_LEVEL_1;
                        break;
                    case 2:
                        securityLevel = UEM_MIFARE_PLUS_SECURITY_LEVEL_2;
                        break;
                    case 3:
                        securityLevel = UEM_MIFARE_PLUS_SECURITY_LEVEL_3;
                        break;
                }
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0xF0);
                command.write(logicalChannel);
                command.write(((authenticationType > UEM_MIFARE_PLUS_AUTHENTICATION_TYPE_FIRST) ? 0x02 : 0x00) | ((securityLevel << 2) & 0x0C));
                command.write(keyType);
                command.write(sector);
                command.write((authenticationType == UEM_MIFARE_PLUS_AUTHENTICATION_TYPE_RESET) ? 0x00 : keyNumber);// ucAuthType == 0x00 - сброс аутентификации
                command.write(keyVersion);
                if (pcdCapabilities == null) {
                    command.write((byte) 0x00);
                } else {
                    command.write((byte) pcdCapabilities.length);
                    if (pcdCapabilities.length > 0) {
                        command.write(pcdCapabilities, 0, pcdCapabilities.length);
                    }
                }
                if (diversificationVector == null) {
                    command.write((byte) 0x00);
                } else {
                    command.write((byte) diversificationVector.length);
                    if (diversificationVector.length > 0) {
                        command.write(diversificationVector, 0, diversificationVector.length);
                    }
                }
                byte[] res = dispatchCommand(command.toByteArray());
                if (res != null && res.length >= 2) {
                    SAM.SW[0] = res[0];
                    SAM.SW[1] = res[1];
                }
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }

            public byte[] readDataSAM(int valueType, int sector, int block, int blocksCount) {
                return readDataSAM(valueType, sector, block, blocksCount, true, true, true);
            }

            public byte[] readDataSAM(int valueType, int sector, int block, int blocksCount, boolean useCommandMAC, boolean useCommandEncryption, boolean useResponseMAC) {
                return readDataSAM((byte)0x00, valueType, sector, block, blocksCount, useCommandMAC, useCommandEncryption, useResponseMAC);
            }

            public byte[] readDataSAM(int logicalChannel, int valueType, int sector, int block, int blocksCount, boolean useCommandMAC, boolean useCommandEncryption, boolean useResponseMAC) {
                byte encryptionMode = (byte)((useCommandMAC?0x00:0x04) | (useCommandEncryption?0x00:0x02) | (useResponseMAC?0x01:0x00));
                byte[] res = dispatchCommand(new byte[] {(byte)0xF4, (byte)logicalChannel, encryptionMode, (byte)valueType, (byte)sector, (byte)block, (byte)blocksCount});
                if (res != null && res.length >= 2) {
                    SAM.SW[0] = res[0];
                    SAM.SW[1] = res[1];
                }
                if (!getLastError().equals(UEM_STATUS_SUCCESS))
                    return null;
                if ((res == null) || (res.length < 4)) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }

                int len = res[2] + res[3] * 256;
                if (res.length != 4 + len) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                    return null;
                }
                return Arrays.copyOfRange(res, 4, res.length);
            }

            public boolean writeDataSAM(int valueType, int sector, int block, int blocksCount, byte[] data) {
                return writeDataSAM(valueType, sector, block, blocksCount, data, true, true, true);
            }

            public boolean writeDataSAM(int valueType, int sector, int block, int blocksCount, byte[] data, boolean useCommandMAC, boolean useCommandEncryption, boolean useResponseMAC) {
                return writeDataSAM((byte)0x00, valueType, sector, block, blocksCount, data, useCommandMAC, useCommandEncryption, useResponseMAC);
            }

            public boolean writeDataSAM(int logicalChannel, int valueType, int sector, int block, int blocksCount, byte[] data, boolean useCommandMAC, boolean useCommandEncryption, boolean useResponseMAC) {
                if (data == null || data.length != 16 * blocksCount) {
                    mLastError = UEM_ERROR_INCORRECT_PARAMETER;
                    return false;
                }
                byte encryptionMode = (byte)((useCommandMAC?0x00:0x04) | (useCommandEncryption?0x00:0x02) | (useResponseMAC?0x01:0x00));
                ByteArrayOutputStream command = new ByteArrayOutputStream();
                command.write(0xF5);
                command.write(logicalChannel);
                command.write(encryptionMode);
                command.write(valueType);
                command.write(sector);
                command.write(block);
                command.write(blocksCount);
                command.write(data, 0, data.length);
                byte[] res = dispatchCommand(command.toByteArray());
                if (res != null && res.length >= 2) {
                    SAM.SW[0] = res[0];
                    SAM.SW[1] = res[1];
                }
                return getLastError().equals(UEM_STATUS_SUCCESS);
            }
        }

        public CommandsMifarePlus MifarePlus = new CommandsMifarePlus();
    }

    public CommandsCards Cards = new CommandsCards();



    /*private byte[] dispatchCommand(Object oReader, byte[] command) {
        String interfaceType = mConnectedReaders.get(oReader);
        if (interfaceType == null) {
            throw new IllegalArgumentException("Reader is not connected!");
        }
        if (interfaceType.equals(UEM_INTERFACE_USB)) {
            byte[] resp = this.USB.Tranceive((UsbDeviceConnection)oReader, command);
            if (resp[0] != command[0]) {
                lastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return null;
            }
            return Arrays.copyOfRange(resp, 0, resp.length-2);
        }
        return null;
    }*/

    public String getInterfaceType() {
        return mInterfaceType;
    }

    private byte[] dispatchCommand(byte[] command) {
        //String interfaceType = mConnectedReaders.get(oReader);
        mLastError = UEM_STATUS_SUCCESS;
        mLastErrorCode = 0x00;
        if (mInterfaceType == null) {
            //throw new IllegalArgumentException("Reader is not connected!");
            mLastError = UEM_ERROR_READER_NOT_CONNECTED;
            return null;
        }
        if (mInterfaceType.equals(UEM_INTERFACE_USB)) {
            /*final*/ byte[] resp = this.USB.transceive(command);
            /*if (resp != null)
                ((Activity)mContext).runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        Toast.makeText(mContext,
                                UemReader.Helper.byteArrayToHexString(resp), Toast.LENGTH_LONG)
                                .show();
                    }});*/
            if (resp == null) {
                mLastError = UEM_ERROR_READER_UNSUCCESSFUL;
                return null;
            }
            if (resp.length < 2 || (resp[0] != command[0])) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return null;
            }
            if (resp[1] != 0x00) {
                mLastError = UEM_ERROR_READER_UNSUCCESSFUL;
                mLastErrorCode = resp[1];
                if (resp.length == 2)
                    return null;
            }
            return Arrays.copyOfRange(resp, 2, resp.length);
        }
        if (mInterfaceType.equals(UEM_INTERFACE_BLUETOOTH)) {
            boolean res = this.Bluetooth.sendCommand(command);
            if (res) {
                mLastError = UEM_STATUS_OPERATION_PENDING;
                return null;
            }
        }
        return null;
    }

    public boolean connect(String readerName) {
        if (this.USB.connect(readerName))
            return true;
        if (!getLastError().equals(UEM_ERROR_READER_INCORRECT_NAME))
            return false;
        return this.Bluetooth.connect(readerName);
    }

    public boolean disconnect() {
        if (mInterfaceType == null) {
            mLastError = UEM_ERROR_READER_NOT_CONNECTED;
            return false;
        }
        if (mInterfaceType.equals(UEM_INTERFACE_USB))
            return this.USB.disconnect();
        if (mInterfaceType.equals(UEM_INTERFACE_BLUETOOTH))
            return this.Bluetooth.disconnect();
        return false;
    }

    public interface BluetoothCallbacks {
        void onBluetoothReaderFound(String readerName);
        void onBluetoothReaderConnected();
        void onBluetoothResponse(byte[] response);
    }

    public interface UsbCallbacks {
        void onUsbReaderAttached(String readerName);
        void onUsbReaderDetached(String readerName);
    }

    public void startReceivingBroadcasts() {
        try {
            this.USB.startReceivingBroadcasts();
            this.Bluetooth.startReceivingBroadcasts();
        } catch (Exception e) {};
    }

    public void stopReceivingBroadcasts() {
        try {
            this.USB.stopReceivingBroadcasts();
            this.Bluetooth.stopReceivingBroadcasts();
        } catch (Exception e) {};
    }

    public class InterfaceUSB {
        private HashMap<String, UsbDevice> mReadersList = new HashMap<String, UsbDevice>();

        private UsbDevice mConnectedReader = null;

        private UsbCallbacks mUsbCallbacks = null;

        private PendingIntent mPermissionIntent = null;

        private BroadcastReceiver mBroadcastReceiver = null;

        private IntentFilter mBroadcastIntentFilter = null;

        private boolean mBroadcastsDisabled = false;

        private UemReader mReader = null;

        InterfaceUSB () {
            /*new Thread(new Runnable() {
                public void run() {

                }
            }).start();*/
        }

        private void initialize(){
            final String ACTION_USB_PERMISSION =
                    "ru.microem.readerlibrary.USB_PERMISSION";
            mBroadcastReceiver = new BroadcastReceiver() {

                public void onReceive(Context context, final Intent intent) {

                    Handler handler = new Handler();
                    handler.postDelayed(new Runnable() {
                        @Override
                        public void run() {
                            if (mBroadcastsDisabled)
                                return;

                            final String action = intent.getAction();
                            Log.d("---", "Received Broadcast: "+action);
                            /*((Activity) mContext).runOnUiThread(new Runnable() {
                                @Override
                                public void run() {
                                    Toast.makeText(mContext,
                                            "Received Broadcast: "+action, Toast.LENGTH_LONG)
                                            .show();
                                }
                            });*/
                            if (ACTION_USB_PERMISSION.equals(action)) {
                                synchronized (this) {
                                    UsbDevice device = (UsbDevice)intent.getParcelableExtra(UsbManager.EXTRA_DEVICE);

                                    if (intent.getBooleanExtra(UsbManager.EXTRA_PERMISSION_GRANTED, false)) {
                                        if(device != null){
                                            //call method to set up device communication
                                            ((Activity) mContext).runOnUiThread(new Runnable() {
                                                @Override
                                                public void run() {
                                                    Toast.makeText(mContext,
                                                            "ACTION_USB_PERMISSION", Toast.LENGTH_LONG)
                                                            .show();
                                                }
                                            });
                                        }
                                    }
                                    else {
                                        Log.d(TAG, "permission denied for device " + device);
                                    }
                                }
                            }
                            if (UsbManager.ACTION_USB_DEVICE_ATTACHED.equals(action) || UsbManager.ACTION_USB_ACCESSORY_ATTACHED.equals(action)) {
                                synchronized (this) {
                                    final UsbDevice device = (UsbDevice)intent.getParcelableExtra(UsbManager.EXTRA_DEVICE);
                                    if (device != null && (device.getVendorId() == 49745) && (device.getProductId() == 4874)) {
                                        final UsbManager usbManager = (UsbManager) mContext.getSystemService(Context.USB_SERVICE);
                                        mReadersList = usbManager.getDeviceList();

                                        //usbManager.requestPermission(device2, mPermissionIntent);

                                        /*final UsbDevice device2 = mReadersList.get(device.getDeviceName());
                                        final Handler handler = new Handler();
                                        handler.postDelayed(new Runnable() {
                                            @Override
                                            public void run() {
                                                if (usbManager.openDevice(device2) == null) {
                                                    ((Activity) mContext).runOnUiThread(new Runnable() {
                                                        @Override
                                                        public void run() {
                                                            Toast.makeText(mContext,
                                                                    "openDevice failed!", Toast.LENGTH_LONG)
                                                                    .show();
                                                        }
                                                    });
                                                }
                                            }
                                        }, 5000);*/
                                       /* if (usbManager.openDevice(device2) == null) {
                                            ((Activity) mContext).runOnUiThread(new Runnable() {
                                                @Override
                                                public void run() {
                                                    Toast.makeText(mContext,
                                                            "openDevice failed!", Toast.LENGTH_LONG)
                                                            .show();
                                                }
                                            });
                                        }*/
                                        //discover();
                                        if (!mReadersList.containsKey(device.getDeviceName())) {
                                            mReadersList.put(device.getDeviceName(), device);
                                            if (mUsbCallbacks != null) {
                                                //new Thread(new Runnable() {
                                                //    public void run() {
                                                //mUsbCallbacks.onUsbReaderAttached(device.getDeviceName());
                                                //    }
                                                //}).start();
                                                /*Handler handler = new Handler();
                                                handler.postDelayed(new Runnable() {
                                                    @Override
                                                    public void run() {
                                                        mUsbCallbacks.onUsbReaderAttached(device.getDeviceName());
                                                    }
                                                }, 10);*/
                                                ((Activity) mContext).runOnUiThread(new Runnable() {
                                                    @Override
                                                    public void run() {
                                                        mUsbCallbacks.onUsbReaderAttached(device.getDeviceName());
                                                    }
                                                });
                                            }
                                        }
                                        Log.d("---", "USB Attached..");
                                    }
                                }
                            }
                            if (UsbManager.ACTION_USB_DEVICE_DETACHED.equals(action) || UsbManager.ACTION_USB_ACCESSORY_DETACHED.equals(action)) {
                                synchronized (this) {
                                    final UsbDevice device = (UsbDevice) intent.getParcelableExtra(UsbManager.EXTRA_DEVICE);
                                    if (device != null && (device.getVendorId() == 49745) && (device.getProductId() == 4874)) {
                                        if (mConnectedReader == device)
                                            disconnect();
                                        //discover();
                                        if (mReadersList.containsKey(device.getDeviceName())) {
                                            mReadersList.remove(device.getDeviceName());
                                            if (mUsbCallbacks != null) {

                                                ((Activity) mContext).runOnUiThread(new Runnable() {
                                                    @Override
                                                    public void run() {
                                                        mUsbCallbacks.onUsbReaderDetached(device.getDeviceName());
                                                    }
                                                });
                                            }
                                            Log.d("---", "USB Detached..");
                                        }
                                    }
                                }
                            }

                        }
                    }, 1000);
                }
            };

            mPermissionIntent = PendingIntent.getBroadcast(mContext, 0, new Intent(ACTION_USB_PERMISSION), 0);

            mBroadcastIntentFilter = new IntentFilter();
            mBroadcastIntentFilter.addAction(UsbManager.ACTION_USB_DEVICE_ATTACHED);
            mBroadcastIntentFilter.addAction(UsbManager.ACTION_USB_DEVICE_DETACHED);
            mBroadcastIntentFilter.addAction(ACTION_USB_PERMISSION);

            discover();
        }

        public void setCallbacks(UsbCallbacks callbacks) {
            this.mUsbCallbacks = callbacks;
        }

        public void startReceivingBroadcasts() {
            if (mBroadcastReceiver != null && mBroadcastIntentFilter != null) {
                mContext.registerReceiver(mBroadcastReceiver, mBroadcastIntentFilter);
                mBroadcastsDisabled = false;
            }

            UsbManager usbManager = (UsbManager)mContext.getSystemService(Context.USB_SERVICE);
            HashMap<String, UsbDevice> currectReadersList = usbManager.getDeviceList();
            for (final UsbDevice usbDevice: currectReadersList.values()) {
                if ((usbDevice.getVendorId() == 49745) && (usbDevice.getProductId() == 4874)) {
                    if (!mReadersList.containsKey(usbDevice.getDeviceName())) {
                        //mReadersList.put(usbDevice.getDeviceName(), usbDevice);
                        if (mUsbCallbacks != null) {
                            /*new Thread(new Runnable() {
                                public void run() {
                                    mUsbCallbacks.onUsbReaderAttached(usbDevice.getDeviceName());
                                }
                            }).start();*/

                            Handler handler = new Handler();
                            handler.postDelayed(new Runnable() {
                                @Override
                                public void run() {
                                    mUsbCallbacks.onUsbReaderAttached(usbDevice.getDeviceName());
                                }
                            }, 10);
                        }
                        Log.d("---", "USB Attached..");
                    }
                }
            }
            for (final UsbDevice usbDevice: mReadersList.values()) {
                if ((usbDevice.getVendorId() == 49745) && (usbDevice.getProductId() == 4874)) {
                    if (!currectReadersList.containsKey(usbDevice.getDeviceName())) {
                        //mReadersList.put(usbDevice.getDeviceName(), usbDevice);
                        if (mUsbCallbacks != null) {
                            /*new Thread(new Runnable() {
                                public void run() {
                                    mUsbCallbacks.onUsbReaderAttached(usbDevice.getDeviceName());
                                }
                            }).start();*/

                            Handler handler = new Handler();
                            handler.postDelayed(new Runnable() {
                                @Override
                                public void run() {
                                    mUsbCallbacks.onUsbReaderDetached(usbDevice.getDeviceName());
                                }
                            }, 10);
                        }
                        Log.d("---", "USB Detached..");
                    }
                }
            }
            mReadersList = currectReadersList;
        }

        public void stopReceivingBroadcasts() {
            if (mBroadcastReceiver != null) {
                mBroadcastsDisabled = true;
                mBroadcastReceiver.abortBroadcast();
                mContext.unregisterReceiver(mBroadcastReceiver);
                //discover();
            }
        }

        public String[] discover() {
            UsbManager usbManager = (UsbManager)mContext.getSystemService(Context.USB_SERVICE);
            mReadersList = usbManager.getDeviceList();
            ArrayList<String> arReaders = new ArrayList<String>();

            for (UsbDevice usbDevice: mReadersList.values()) {
                if ((usbDevice.getVendorId() == 49745) && (usbDevice.getProductId() == 4874)) {
                    arReaders.add(usbDevice.getDeviceName());
                    //if (mDiscoveryCallback != null)
                    //    mDiscoveryCallback.onNewReaderFound(usbDevice.getDeviceName());
                }
            }

            Object[] objReaders = arReaders.toArray();
            String[] strReaders = Arrays.copyOf(objReaders, objReaders.length, String[].class);
            return strReaders;
        }

        private UsbEndpoint mEndpointBulkIn = null;
        private UsbEndpoint mEndpointBulkOut = null;

        /*public UsbDeviceConnection connect(String readerName) {
            if (!readersList.containsKey(readerName)) {
                lastError = UEM_ERROR_READER_INCORRECT_NAME;
                return null;
            }

            UsbManager usbManager = (UsbManager) getSystemService(Context.USB_SERVICE);
            UsbDevice reader = readersList.get(readerName);
            UsbInterface interfaceUEM = reader.getInterface(0);
            endpointBulkIn = interfaceUEM.getEndpoint(0);
            endpointBulkOut = interfaceUEM.getEndpoint(1);
            UsbDeviceConnection connection = usbManager.openDevice(reader);
            if (connection == null) {
                lastError = UEM_ERROR_READER_CONNECTION_FAILED;
                return null;
            }

            connection.claimInterface(interfaceUEM, true);
            mConnectedReaders.put(connection, UEM_INTERFACE_USB);
            interfaceType = UEM_INTERFACE_USB;

            return connection;
        }*/

        private boolean connect(final String readerName) {
            if (mConnection != null) {
                mLastError = UEM_ERROR_READER_ALREADY_CONNECTED;
                return false;
            }
            if (!mReadersList.containsKey(readerName)) {
                mLastError = UEM_ERROR_READER_INCORRECT_NAME;
                return false;
            }

            UsbManager usbManager = (UsbManager)mContext.getSystemService(Context.USB_SERVICE);
            UsbDevice reader = mReadersList.get(readerName);

            //usbManager.requestPermission(reader, mPermissionIntent);

            UsbInterface interfaceUEM = reader.getInterface(0);
            mEndpointBulkIn = interfaceUEM.getEndpoint(0);
            mEndpointBulkOut = interfaceUEM.getEndpoint(1);
            UsbDeviceConnection connection = usbManager.openDevice(reader);
            if (connection == null) {
                mLastError = UEM_ERROR_READER_CONNECTION_FAILED;
                return false;
            }

            connection.claimInterface(interfaceUEM, true);
            mConnection = connection;
            mInterfaceType = UEM_INTERFACE_USB;
            mConnectedReader = reader;

            return true;
        }

        /*public boolean disconnect(UsbDeviceConnection connection) {
            if (connection == null)
                return false;

            connection.close();
            mConnectedReaders.remove(connection);
            connection = null;

            return true;
        }*/

        private boolean disconnect() {
            if (mConnection == null)
                return false;

            ((UsbDeviceConnection)mConnection).close();
            mConnection = null;
            mInterfaceType = null;
            mConnectedReader = null;

            return true;
        }

        /*public byte[] Tranceive(UsbDeviceConnection connection, byte[] sendBuffer) {
            if (connection == null || sendBuffer == null)
                throw new NullPointerException("Parameter is null!");
            if (sendBuffer.length == 0)
                throw new IllegalArgumentException("Incorrect parameter!");

            int TIMEOUT = 0;
            connection.bulkTransfer(endpointBulkOut, sendBuffer, sendBuffer.length, TIMEOUT);

            byte[] receiveBuffer = new byte[256];
            final int nReceived = connection.bulkTransfer(endpointBulkIn, receiveBuffer, receiveBuffer.length, TIMEOUT);

            if (nReceived < 0) {
                lastError = UEM_ERROR_READER_RESPONSE_FAILURE;
                return null;
            }

            byte[] response = UemReader.Helper.unwrapResponse(Arrays.copyOf(receiveBuffer, nReceived));

            return response;
        }*/

        private byte[] transceive(byte[] sendBuffer) {
            if (mConnection == null || sendBuffer == null)
                throw new NullPointerException("Parameter is null!");

            if (sendBuffer.length == 0)
                throw new IllegalArgumentException("Incorrect parameter!");

            int TIMEOUT = 0;
            byte[] command = UemReader.Helper.wrapCommand(sendBuffer, mReader);
            if (command == null)
                throw new NullPointerException("Wrapped command is null!");
            ((UsbDeviceConnection)mConnection).bulkTransfer(mEndpointBulkOut, command, command.length, TIMEOUT);

            byte[] receiveBuffer = new byte[256];
            final int nReceived = ((UsbDeviceConnection)mConnection).bulkTransfer(mEndpointBulkIn, receiveBuffer, receiveBuffer.length, TIMEOUT);

            if (nReceived < 0) {
                mLastError = UEM_ERROR_READER_RESPONSE_FAILURE;
                return null;
            }

            byte[] response = UemReader.Helper.unwrapResponse(Arrays.copyOf(receiveBuffer, nReceived), mReader);
            if (response == null) {
                mLastError = UEM_ERROR_READER_RESPONSE_FAILURE;
                return null;
            }
            return response;
        }
    }

    public InterfaceUSB USB = new InterfaceUSB();

    public class InterfaceBluetooth {
        private BluetoothAdapter mBluetoothAdapter = null;
        private BluetoothLeScanner mBluetoothLeScanner = null;
        private BluetoothManager mBluetoothManager = null;
        private BluetoothScanCallback  mBluetoothScanCallback = null;
        private BroadcastReceiver mBroadcastReceiver = null;
        IntentFilter mFilterBluetoothStateChanged = null;
        //private BluetoothGatt gattClient;

        private BluetoothGattCharacteristic mCharacteristicFIFO;
        private BluetoothGattCharacteristic mCharacteristicCREDITS;

        final UUID SERVICE_UUID_SERIAL = UUID.fromString("2456e1b9-26e2-8f83-e744-f34f01e9d701");
        final UUID CHARACTERISTIC_UUID_FIFO = UUID.fromString("2456e1b9-26e2-8f83-e744-f34f01e9d703");
        final UUID CHARACTERISTIC_UUID_CREDITS = UUID.fromString("2456e1b9-26e2-8f83-e744-f34f01e9d704");

        int mFinalMTU = 0;

        byte mLastCommandCode = 0x00;

        boolean mScanAfterBTEnable = false;

        private BluetoothCallbacks mBluetoothCallbacks = null;

        private UemReader mReader = null;

        InterfaceBluetooth() {
            mBluetoothScanCallback = new BluetoothScanCallback();
            mBroadcastReceiver = new BroadcastReceiver() {
                @Override
                public void onReceive(Context context, Intent intent) {
                    Object action = intent.getAction();
                    if (action == BluetoothAdapter.ACTION_STATE_CHANGED) {
                        int state = intent.getIntExtra(BluetoothAdapter.EXTRA_STATE, BluetoothAdapter.ERROR);
                        switch (state) {
                            case BluetoothAdapter.STATE_ON: {
                                if (mScanAfterBTEnable) {
                                    mScanAfterBTEnable = false;
                                    startScanner();
                                }
                            }
                            case BluetoothAdapter.STATE_OFF: {
                                //bluetoothAdapter.enable();
                            }
                        }
                    }
                }
            };
        }

        public void setCallbacks(BluetoothCallbacks callbacks) {
            this.mBluetoothCallbacks = callbacks;
        }

        public void startReceivingBroadcasts() {
            if (mBroadcastReceiver != null && mFilterBluetoothStateChanged != null)
                mContext.registerReceiver(mBroadcastReceiver, mFilterBluetoothStateChanged);
        }

        public void stopReceivingBroadcasts() {
            if (mBroadcastReceiver != null)
                mContext.unregisterReceiver(mBroadcastReceiver);

        }

        private void startScanner() {
            if (!mBluetoothAdapter.isEnabled())
                return;
            if (mBluetoothLeScanner == null)
                mBluetoothLeScanner = mBluetoothAdapter.getBluetoothLeScanner();

            ScanFilter.Builder builder = new ScanFilter.Builder();
                            /*switch(scanType) {
                                case SCAN_BY_NAME:
                                    builder.setDeviceName(data);
                                    break;
                            }*/
            Vector<ScanFilter> filter = new Vector<ScanFilter>();
            filter.add(builder.build());
            ScanSettings.Builder builderScanSettings = new ScanSettings.Builder();
            //builderScanSettings.setScanMode(ScanSettings.SCAN_MODE_BALANCED);
            //builderScanSettings.setScanMode(ScanSettings.SCAN_MODE_LOW_POWER);
            builderScanSettings.setScanMode(ScanSettings.SCAN_MODE_LOW_LATENCY);
            builderScanSettings.setReportDelay(0);
            mBluetoothLeScanner.stopScan(mBluetoothScanCallback);
            mBluetoothLeScanner.flushPendingScanResults(mBluetoothScanCallback);
            mBluetoothLeScanner.startScan(filter, builderScanSettings.build(),mBluetoothScanCallback);
            //bluetoothLeScanner.startScan(bluetoothScanCallback);
        }

        private void checkPermissions() {
            if(Build.VERSION.SDK_INT > Build.VERSION_CODES.LOLLIPOP){
                int permissionCheck = mContext.checkSelfPermission("Manifest.permission.ACCESS_FINE_LOCATION");
                permissionCheck += mContext.checkSelfPermission("Manifest.permission.ACCESS_COARSE_LOCATION");
                if (permissionCheck != 0) {

                    ((Activity)mContext).requestPermissions(new String[]{Manifest.permission.ACCESS_FINE_LOCATION, Manifest.permission.ACCESS_COARSE_LOCATION}, 1001); //Any number
                }
            }else{
                Log.d(TAG, "checkBTPermissions: No need to check permissions. SDK version < LOLLIPOP.");
            }
        }

        private void initialize() {
            checkPermissions();
            mBluetoothManager = (BluetoothManager)mContext.getSystemService(BLUETOOTH_SERVICE);
            mBluetoothAdapter = BluetoothAdapter.getDefaultAdapter();//bluetoothManager.getAdapter();

            //bluetoothScanCallback = new BluetoothScanCallback();
            mFilterBluetoothStateChanged = new IntentFilter(BluetoothAdapter.ACTION_STATE_CHANGED);
        }

        public boolean turnRadioOn() {
            if (mBluetoothAdapter.isEnabled())
                return true;

            mScanAfterBTEnable = true;

            Intent enableIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
            //startActivityForResult(enableIntent, REQUEST_ENABLE_BLUETOOTH);
            //Intent discoverableIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_DISCOVERABLE);
            //discoverableIntent.putExtra(BluetoothAdapter.EXTRA_DISCOVERABLE_DURATION, 60);
            //startActivity(discoverableIntent);
            ((Activity)mContext).startActivityForResult(enableIntent, 0);

            return false;
        }

        public void turnRadioOff() {
            if (mBluetoothAdapter.isEnabled()) {
                mBluetoothAdapter.disable();
            }
        }

        public void startDiscovery() {
            Log.i(TAG,"startScan()");
            mBleDevices.clear();
            if (mBluetoothManager == null) {
                initialize();
                //((Activity)mContext).registerReceiver(broadcastReceiver, filterBluetoothStateChanged);
                //bluetoothAdapter.disable();
            } else {
                stopDiscovery();
                //((Activity)mContext).registerReceiver(broadcastReceiver, filterBluetoothStateChanged);
            }
            if (turnRadioOn())
                startScanner();
        }

        public void stopDiscovery() {
            //disconnect();
            if (mBluetoothLeScanner != null) {
                try {
                    ((Activity) mContext).unregisterReceiver(mBroadcastReceiver);
                } catch (Exception e) {};
                mBluetoothLeScanner.stopScan(mBluetoothScanCallback); // stop scan
                mBluetoothLeScanner.flushPendingScanResults(mBluetoothScanCallback);
                //bluetoothLeScanner = null;
            }
            //bluetoothManager = null;
            //bluetoothAdapter = null;
            //filterBluetoothStateChanged = null;
        }

        private boolean connect(String readerName) {
            if (mConnection != null) {
                mLastError = UEM_ERROR_READER_ALREADY_CONNECTED;
                return false;
            }
            if (!mBleDevices.containsKey(readerName)) {
                mLastError = UEM_ERROR_READER_INCORRECT_NAME;
                return false;
            }

            return connectDevice(mBleDevices.get(readerName));
        }

        // BLUETOOTH CONNECTION
        private boolean connectDevice(BluetoothDevice device) {
            if (device == null) Log.i(TAG,"Device is null");
            GattClientCallback gattClientCallback = new GattClientCallback();
            //gattClient = device.connectGatt(this,false,gattClientCallback);

            if (((BluetoothGatt)mConnection) != null && mBluetoothManager.getConnectedDevices(BluetoothProfile.GATT).size() > 0)
                return false;

            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                mConnection = device.connectGatt(mContext, false, gattClientCallback, TRANSPORT_LE,PHY_LE_CODED_MASK ^ PHY_LE_1M_MASK);
            } else if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                mConnection = device.connectGatt(mContext, false, gattClientCallback, TRANSPORT_LE);
            } else {
                mConnection = device.connectGatt(mContext, false, gattClientCallback);
            }

            mInterfaceType = UEM_INTERFACE_BLUETOOTH;

            return true;
        }

        private boolean disconnect() {
            if (mConnection == null)
                return false;

            ((BluetoothGatt)mConnection).disconnect();
            ((BluetoothGatt)mConnection).close();
            mConnection = null;
            mInterfaceType = null;

            return true;
        }

        private HashMap<String, BluetoothDevice> mBleDevices = new HashMap<String, BluetoothDevice>();

        // BLE Scan Callbacks
        private class BluetoothScanCallback extends ScanCallback {

            @Override
            public void onScanResult(int callbackType, ScanResult result) {
                Log.i(TAG, "onScanResult");
                if (result.getDevice().getName() != null){
                    if (result.getDevice().getName().contains("ANNA-") || result.getDevice().getName().contains("uEM Reader")) {
                        if(!mBleDevices.containsKey(result.getDevice().getName())) {
                            mBleDevices.put(result.getDevice().getName(), result.getDevice());
                            /*ListView listBLE = findViewById(R.id.bleList);
                            _listBLE.add(result.getDevice().getName());
                            ((ArrayAdapter<String>) listBLE.getAdapter()).notifyDataSetChanged();*/
                            if (mBluetoothCallbacks != null)
                                mBluetoothCallbacks.onBluetoothReaderFound(result.getDevice().getName());
                        }
                        // When find your device, connect.
                        //connectDevice(result.getDevice());
                        //bluetoothLeScanner.stopScan(bluetoothScanCallback); // stop scan
                    }
                }
            }

            @Override
            public void onBatchScanResults(List<ScanResult> results) {
                Log.i(TAG, "onBathScanResults");

                //bluetoothLeScanner.startScan(bluetoothScanCallback);
            }

            @Override
            public void onScanFailed(int errorCode) {
                Log.i(TAG, "ErrorCode: " + errorCode);
            }
        }

        // Bluetooth GATT Client Callback
        private class GattClientCallback extends BluetoothGattCallback {
            @Override
            public void onMtuChanged(BluetoothGatt gatt, int mtu, int status) {
                super.onMtuChanged(gatt, mtu, status);
                mFinalMTU = mtu;
                gatt.discoverServices();
            }

            @Override
            public void onConnectionStateChange(BluetoothGatt gatt, int status, int newState) {
                super.onConnectionStateChange(gatt, status, newState);
                Log.i(TAG,"onConnectionStateChange");

                if (status == BluetoothGatt.GATT_FAILURE) {
                    Log.i(TAG, "onConnectionStateChange GATT FAILURE");
                    return;
                } else if (status != BluetoothGatt.GATT_SUCCESS) {
                    /*final int s = status;
                    ((Activity)mContext).runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            Toast.makeText(mContext,
                                    "Failed to change connection state... "+String.valueOf(s), Toast.LENGTH_LONG)
                                    .show();
                        }});*/
                    Log.i(TAG, "onConnectionStateChange != GATT_SUCCESS");
                    return;
                }

                if (newState == BluetoothProfile.STATE_CONNECTED) {
                    Log.i(TAG, "onConnectionStateChange CONNECTED");
                    /*((Activity)mContext).runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            Toast.makeText(mContext,
                                    "BLE Device connected!", Toast.LENGTH_LONG)
                                    .show();
                        }});*/
                    gatt.requestMtu(512);
                } else if (newState == BluetoothProfile.STATE_DISCONNECTED) {
                    Log.i(TAG, "onConnectionStateChange DISCONNECTED");
                    /*((Activity)mContext).runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            Toast.makeText(mContext,
                                    "BLE Device disconnected!", Toast.LENGTH_LONG)
                                    .show();
                        }});*/
                }
            }

            @Override
            public void onServicesDiscovered(BluetoothGatt gatt, int status) {
                super.onServicesDiscovered(gatt, status);
                Log.i(TAG,"onServicesDiscovered");
                if (status != BluetoothGatt.GATT_SUCCESS) return;

                // Reference your UUIDs
                mCharacteristicFIFO = gatt.getService(SERVICE_UUID_SERIAL).getCharacteristic(CHARACTERISTIC_UUID_FIFO);
                mCharacteristicCREDITS = gatt.getService(SERVICE_UUID_SERIAL).getCharacteristic(CHARACTERISTIC_UUID_CREDITS);

                String UUID_MASK = "0000%s-0000-1000-8000-00805f9b34fb";
                String CLIENT_CHARACTERISTIC_CONFIG = String.format(UUID_MASK, "2902");

                if ((mCharacteristicFIFO.getProperties() & BluetoothGattCharacteristic.PROPERTY_NOTIFY) > 0) {
                    BluetoothGattDescriptor descriptor = mCharacteristicFIFO.getDescriptor(UUID.fromString(CLIENT_CHARACTERISTIC_CONFIG));
                    if (descriptor == null || gatt == null) return;

                    // Set write type to WRITE_TYPE_DEFAULT to fix an issue seen on some phone models and
                    // Android versions that enable notification with a write_command message instead of a
                    // write_request message (as specified by the Bluetooth specification).
                    mCharacteristicFIFO.setWriteType(BluetoothGattCharacteristic.WRITE_TYPE_DEFAULT);
                    descriptor.setValue(BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE);
                    gatt.setCharacteristicNotification(mCharacteristicFIFO, true);
                    gatt.writeDescriptor(descriptor);
                }
            }

            @Override
            public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
                super.onCharacteristicRead(gatt, characteristic, status);
                Log.i(TAG,"onCharacteristicRead");
            }

            @Override
            public void onCharacteristicWrite(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
                super.onCharacteristicWrite(gatt, characteristic, status);
                Log.i(TAG,"onCharacteristicWrite");
            }

            @Override
            public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
                super.onCharacteristicChanged(gatt, characteristic);
                Log.i(TAG,"onCharacteristicChanged");
                // Here you can read the characteristc's value
                /*final BluetoothGattCharacteristic ch = characteristic;
                ((Activity)mContext).runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        TextView textView = (TextView) findViewById(R.id.outtext);
                        String out = new String(ch.getValue());
                        try {
                            out = bytesToHex(unwrapResponse(ch.getValue()));
                        } catch (Exception e) {
                            System.out.println("error " + e.getMessage());
                            return;
                        }
                        textView.setText(out);
                    }
                });*/

                if (characteristic != mCharacteristicFIFO) {
                    return;
                }

                mLastError = UEM_STATUS_SUCCESS;

                byte[] response = UemReader.Helper.unwrapResponse(characteristic.getValue(), mReader);

                if (response == null || response.length <= 0) {
                    mLastError = UEM_ERROR_READER_RESPONSE_FAILURE;
                } else if (response[0] != mLastCommandCode) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                } else if (response.length == 1) {
                    mLastError = UEM_ERROR_READER_RESPONSE_FAILURE;
                } else if (response[1] != 0x00) {
                    mLastError = UEM_ERROR_READER_UNSUCCESSFUL;
                    mLastErrorCode = response[1];
                }

                response = Arrays.copyOfRange(response, 1, response.length);

                mBluetoothCallbacks.onBluetoothResponse(response);
            }

            @Override
            public void onDescriptorRead(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
                super.onDescriptorRead(gatt, descriptor, status);
                Log.i(TAG,"onDescriptorRead");
            }

            @Override
            public void onDescriptorWrite(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
                super.onDescriptorWrite(gatt, descriptor, status);
                Log.i(TAG,"onDescriptorWrite");

                mBluetoothCallbacks.onBluetoothReaderConnected();
            }

        }

        private boolean sendCommand(byte[] sendBuffer) {
            if (sendBuffer == null)
                throw new NullPointerException("Parameter is null!");

            if (sendBuffer.length == 0)
                throw new IllegalArgumentException("Incorrect parameter!");

            if (((BluetoothGatt)mConnection) == null) {
                mLastError = UEM_ERROR_READER_NOT_CONNECTED;
                return false;
            }

            if (mCharacteristicFIFO == null || (mCharacteristicFIFO.getProperties() & BluetoothGattCharacteristic.PROPERTY_WRITE) <= 0) {
                mLastError = UEM_ERROR_READER_CONNECTION_FAILED;
                return false;
            }

            mLastCommandCode = sendBuffer[0];

            byte[] rawCommand = UemReader.Helper.wrapCommand(sendBuffer, mReader);
            if (rawCommand == null)
                throw new NullPointerException("Command is null!");
            mCharacteristicFIFO.setWriteType(BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE);
            mCharacteristicFIFO.setValue(rawCommand);
            ((BluetoothGatt)mConnection).writeCharacteristic(mCharacteristicFIFO);

            return true;
        }
    }

    public InterfaceBluetooth Bluetooth = new InterfaceBluetooth();

    public static class Helper {
        public static byte[] hexStringToByteArray(String s) {
            s = s.replaceAll("\\s+","");
            int len = s.length();
            byte[] data = new byte[len / 2];

            for (int i = 0; i < len; i += 2)
                data[i / 2] = (byte) ((Character.digit(s.charAt(i), 16) << 4) + Character.digit(s.charAt(i + 1), 16));

            return data;
        }

        final private static char[] hexArray = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

        public static String byteArrayToHexString(byte[] bytes) {
            char[] hexChars = new char[bytes.length * 2];
            int v;

            for (int j = 0; j < bytes.length; j++) {
                v = bytes[j] & 0xFF;
                hexChars[j * 2] = hexArray[v >>> 4];
                hexChars[j * 2 + 1] = hexArray[v & 0x0F];
            }

            return new String(hexChars);
        }

        private static byte[] crc16(byte[] buf, int start, int count) {
            int crc = 0xFFFF;

            for (int pos = start; pos < start + count; pos++) {
                crc ^= (int) buf[pos] & 0xFF;   // XOR byte into least sig. byte of crc

                for (int i = 8; i != 0; i--) {    // Loop over each bit
                    if ((crc & 0x0001) != 0) {      // If the LSB is set
                        crc >>= 1;                    // Shift right and XOR 0x8408
                        crc ^= 0x8408;
                    } else                            // Else LSB is not set
                        crc >>= 1;                    // Just shift right
                }
            }

            return new byte[]{(byte) ((crc ^ 0xFFFF) & 0xFF), (byte) (((crc ^ 0xFFFF) >> 8) & 0xFF)};
        }

        private static byte[] byteStuff(byte[] data) {
            ByteArrayOutputStream stuffedData = new ByteArrayOutputStream();
            for (byte dataByte : data) {
                if ((dataByte & 0xFF) < 0xFD) {
                    stuffedData.write(dataByte);
                } else {
                    stuffedData.write(0xFF);
                    stuffedData.write(0xFF - (dataByte & 0xFF));
                }
            }
            return stuffedData.toByteArray();
        }

        private static byte[] unByteStuff(byte[] rawData) {
            ByteArrayOutputStream unStuffedData = new ByteArrayOutputStream();
            boolean fInvertNext = false;
            for (byte dataByte : rawData) {
                if ((dataByte & 0xFF) == 0xFF) {
                    fInvertNext = true;
                    continue;
                }
                if (fInvertNext) {
                    unStuffedData.write(0xFF - (dataByte & 0xFF));
                    fInvertNext = false;
                } else {
                    unStuffedData.write(dataByte & 0xFF);
                }
            }
            return unStuffedData.toByteArray();
        }

        private static byte nCommand = 0x00;

        private static byte[] wrapCommand(byte[] data) {
            return wrapCommand(data, null);
        }

        private static byte[] wrapCommand(byte[] data, UemReader reader) {
            ByteArrayOutputStream rawData = new ByteArrayOutputStream();

            rawData.write(0x00);
            rawData.write(nCommand++);
            if ((reader != null) && reader.Reader.encryptedMode) {
                rawData.write(0x00);
                data = AES.encryptChannel(data, reader);
                if (data == null)
                    return null;
            }
            rawData.write(data, 0, data.length);

            byte[] FCS = crc16(rawData.toByteArray(), 0, rawData.size());
            rawData.write(FCS, 0, FCS.length);

            byte[] tmpData = byteStuff(rawData.toByteArray());
            rawData = new ByteArrayOutputStream();
            rawData.write(0xFD);
            rawData.write(tmpData, 0, tmpData.length);
            rawData.write(0xFE);

            return rawData.toByteArray();
        }

        private static byte[] unwrapResponse(byte[] rawData) {
            return unwrapResponse(rawData, null);
        }

        private static byte[] unwrapResponse(byte[] rawData, final UemReader reader) {
            rawData = unByteStuff(rawData);
            if ((rawData[0] & 0xFF) != 0xFD)
                return null;
            if ((rawData[rawData.length - 1] & 0xFF) != 0xFE)
                return null;
            byte[] FCS = crc16(rawData, 1, rawData.length - 4);
            if ((FCS[0] & 0xFF) != (rawData[rawData.length - 3] & 0xFF))
                return null;
            if  ((FCS[1] & 0xFF) != (rawData[rawData.length - 2] & 0xFF))
                return null;

            byte[] data = Arrays.copyOfRange(rawData, 3, rawData.length - 3);
            if ((reader != null) && reader.Reader.encryptedMode && (data[0] == 0x00)) {
                data = AES.decryptChannel(Arrays.copyOfRange(data, 1, data.length), reader);
            }
            return data;
        }

        private static byte getActualBlockAddress(int sector, int block) {
            int addr = 4 * sector + block;
            if (sector >= 32)
                addr += 12 * (sector - 32);
            return (byte)addr;
        }

        public static class CRC32 {
            private int mCrc;
            private static boolean mTableBuilt = false;
            private static int[] mTable = new int[0x100];
            private static int POLYNOMIAL = 0x04C11DB7;

            public CRC32() {
                Reset();
            }

            public CRC32(byte[] buf, int siz) {
                Reset();
                Hash(buf, siz);
            }

            public int Get() {
                return mCrc;
            }

            public void Reset() {
                mCrc = ~0;
            }

            public void Hash(byte[] buf, int siz) {
                if(!mTableBuilt)
                    BuildTable();

                for(int i = 0; i < siz; ++i)
                    mCrc = ((mCrc >> 8) & 0x00FFFFFF) ^ mTable[ (mCrc ^ buf[i]) & 0xFF ];
            }

            private static void BuildTable() {
                int i, j;
                for(i = 0; i < 0x100; ++i) {
                    mTable[i] = Reflect((int)i,8) << 24;

                    for(j = 0; j < 8; ++j)
                        mTable[i] = (mTable[i] << 1) ^ ( ((mTable[i] & ((int)((long)1 << 31))) != 0)  ? POLYNOMIAL : 0);

                    mTable[i] = Reflect(mTable[i],32);
                }
                mTableBuilt = true;
            }

            private static int Reflect(int v, int bits) {
                int ret = 0;
                int i;

                --bits;
                for(i = 0; i <= bits; ++i) {
                    if((v & ((int)1 << i)) != 0)
                        ret |= (int)1 << (bits-i);
                }
                return ret;
            }
        };

        private static class AES {
            private static byte[] encryptChannel(byte[] input, UemReader reader) {
                if (input == null || input.length == 0)
                    return null;

                int size = input.length;

                byte[] tmpInp = new byte[256];

                Helper.CRC32 ccc32 = new Helper.CRC32();

                System.arraycopy(input, 0, tmpInp, 0, size);

                ccc32.Hash(tmpInp, size);
                int crc32 = ccc32.Get();

                tmpInp[size++] = (byte)(crc32 & 0xFF);
                tmpInp[size++] = (byte)((crc32 >> 8) & 0xFF);
                tmpInp[size++] = (byte)((crc32 >> 16) & 0xFF);
                tmpInp[size++] = (byte)((crc32 >> 24) & 0xFF);

                tmpInp[size++] = (byte)0x80;
                size = 16*(int)Math.ceil((double)size/16);

                byte[] output = AES.aes128EncryptCbc(tmpInp, size, reader.Reader.sessionKey, reader.Reader.sessionIV, null);
                System.arraycopy(AES.aes128EncryptCbcIvResult, 0, reader.Reader.sessionIV, 0, 16);

                return output;
            }

            private static byte[] decryptChannel(byte[] input, UemReader reader) {
                if (input == null || input.length == 0)
                    return null;

                int size = input.length;

                byte[] tmpInp = new byte[256];

                System.arraycopy(input, 0, tmpInp, 0, size);

                size = 16*(int)Math.ceil((double)size/16);
                byte[] output = AES.aes128DecryptCbc(tmpInp, size, reader.Reader.sessionKey, reader.Reader.sessionIV, null);
                System.arraycopy(AES.aes128DecryptCbcIvResult, 0, reader.Reader.sessionIV, 0, 16);

                for (size = size - 1; size > 0; size--)
                    if (output[size] == (byte)0x80)
                        break;

                if (size <= 4)
                    return null;

                size -= 4;

                Helper.CRC32 ccc32 = new Helper.CRC32();
                ccc32.Hash(output, size);
                int crc32 = ccc32.Get();
                byte[] testCRC = new byte[]{(byte)(crc32 & 0xFF), (byte)((crc32 >> 8) & 0xFF), (byte)((crc32 >> 16) & 0xFF), (byte)((crc32 >> 24) & 0xFF)};

                //reader.debugData = Arrays.copyOfRange(output, size, size + 4);
                if (!Arrays.equals(Arrays.copyOfRange(output, size, size + 4), testCRC))
                    return null;

                return Arrays.copyOfRange(output, 0, size);
            }

            private static byte[] aes128CbcEnMAC(byte[] pInp, int ByteSize, byte[] pKey, byte[] pOut ) {
                // Parameters: BYTE* pInp, int ByteSize, BYTE* pKey, BYTE* pOut
                int i, j, BlockQty;
                byte[] pPlain;
                byte[] pSubKey;
                byte[] wrk = new byte[]{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
                byte[] pK1 = new byte[16];
                byte[] pK2 = new byte[16];

                byte[][] K12 = generateSubkey( pKey );
                pK1 = K12[0];
                pK2 = K12[1];

                if( ((byte)ByteSize & 0x0f) != 0x00) {
                    pSubKey = pK2;
                    pInp[ByteSize] = (byte)0x80;
                    for( i = ByteSize + 1; ( i & 0x0f ) > 0; i++ )
                        pInp[i] = 0;
                }
                else
                    pSubKey = pK1;
                BlockQty = ( ByteSize + 15 ) >> 4;
                //pPlain = pInp;
                for( i = 0; i < BlockQty; i++ ) {
                    pPlain = Arrays.copyOfRange(pInp, 16 * i, 16 * (i + 1));
                    for( j = 0; j < 16; j++ ) {
                        wrk[j] ^= pPlain[j];
                        if( i == BlockQty - 1 )
//                wrk[j] ^= pK2[j];
                            wrk[j] ^= pSubKey[j];
                    }
                    pOut = AES.aes128Encrypt16( wrk, pKey );
                    wrk = Arrays.copyOf(pOut, 16);
                    //memcpy( wrk, pOut, 16 );
                    //pPlain += 16;
                }
                return pOut;
            }

            private static byte[] aes192CbcEnMac(byte[] pInp, int BlockQty, byte[] pKey, byte[] pOut ) {
                return aes192CbcEnMac( pInp, BlockQty, pKey,  pOut, null );
            }

            private static byte[] aes192CbcEnMac(byte[] pInp, int BlockQty, byte[] pKey, byte[] pOut, byte[] pIV ) {
                int i;
                byte[] pPlain = null;
                byte[] wrk = new byte[24];
                Arrays.fill(wrk, (byte)0x00);
                byte[] pK1 = new byte[24];
                byte[] pK2 = new byte[24];
                //pPlain = pInp;

                byte[][] K12 = generateSubkey(pKey);
                pK1 = K12[0];
                pK2 = K12[1];

                if (pIV != null)
                    wrk = xor192(wrk, pIV);

                for( i = 0; i < BlockQty - 1; i++ ) {
                    pPlain = Arrays.copyOfRange(pInp, 24 * i, 24 * (i + 1));
                    wrk = xor192(wrk, pPlain);
                    pOut = aes192Encrypt24( wrk, pKey );
                    wrk = Arrays.copyOf(pOut, 24);
                    //memcpy( wrk, pOut, 24 );
                    //pPlain += 24;
                }

                wrk = xor192(wrk, pPlain);
                wrk = xor192(wrk, pK2);
                pOut = aes192Encrypt24( wrk, pKey );

                return pOut;
            }

            //количество столбцов в матрице, используемой для шифрования
            private final static int NB = 4;
            //количество столбцов в матрице, используемой для представления ключа
            private final static int NK = 4;
            //количество рауднов шифрования
            private final static int NR = 10;

            //шифрование данных input[16] ключом key[16],
            //результат сохраняется в output[16]
            private static byte[] aes128Encrypt16(byte[] input, byte[] key) {
                byte[] output = new byte[3 + 4 * (NB - 1) + 1];
                byte[][] state = new byte[4][NB];
                byte[][] w = new byte[4][NB * (NR + 1)];
                int r, c, round;

                for (r = 0; r <= 3; r++)
                    for (c = 0; c <= NB - 1; c++)
                        state[r][c] = input[r + 4 * c];

                w = KeyExpansion(key, w);
                state = addRoundKey(state, w, 0);

                for (round = 1; round <= NR - 1; round++) {
                    state = subBytes(state);
                    state = shiftRows(state);
                    state = mixColumns(state);
                    state = addRoundKey(state, w, round * NB);
                }

                state = subBytes(state);
                state = shiftRows(state);
                state = addRoundKey(state, w, NR * NB);

                for (r = 0; r <= 3; r++)
                    for (c = 0; c <= NB - 1; c++)
                        output[r + 4 * c] = state[r][c];

                return output;
            }

            //шифрование данных input[24] ключом key[24],
            //результат сохраняется в output[24]
            private static byte[] aes192Encrypt24(byte[] input, byte[] key) {
                byte[] output = new byte[3 + 4 * (NB - 1) + 1];
                byte[][] state = new byte[4][NB];
                byte[][] w = new byte[4][NB * (NR + 1)];
                int r, c, round;

                for (r = 0; r <= 3; r++)
                    for (c = 0; c <= NB - 1; c++)
                        state[r][c] = input[r + 4 * c];

                w = KeyExpansion(key, w);
                state = addRoundKey(state, w, 0);

                for (round = 1; round <= NR - 1; round++) {
                    state = subBytes(state);
                    state = shiftRows(state);
                    state = mixColumns(state);
                    state = addRoundKey(state, w, round * NB);
                }

                state = subBytes(state);
                state = shiftRows(state);
                state = addRoundKey(state, w, NR * NB);

                for (r = 0; r <= 3; r++)
                    for (c = 0; c <= NB - 1; c++)
                        output[r + 4 * c] = state[r][c];

                return output;
            }

            //расшифровка данных input[16] ключом key[16],
            //результат сохраняется в output[16]
            private static byte[] aes128Decrypt16(byte[] input, byte[] key) {
                byte[] output = new byte[3 + 4 * (NB - 1) + 1];
                byte[][] state = new byte[4][NB];
                byte[][] w = new byte[4][NB * (NR + 1)];
                int r, c, round;

                for (r = 0; r <= 3; r++)
                    for (c = 0; c <= NB - 1; c++)
                        state[r][c] = input[r + 4 * c];

                w = KeyExpansion(key, w);
                state = addRoundKey(state, w, NR * NB);

                for (round = NR - 1; round >= 1; round--) {
                    state = invShiftRows(state);
                    state = invSubBytes(state);
                    state = addRoundKey(state, w, round * NB);
                    state = invMixColumns(state);
                }

                state = invShiftRows(state);
                state = invSubBytes(state);
                state = addRoundKey(state, w, 0);

                for (r = 0; r <= 3; r++)
                    for (c = 0; c <= NB - 1; c++)
                        output[r + 4 * c] = state[r][c];

                return output;
            }

            //расшифровка данных input[24] ключом key[24],
            //результат сохраняется в output[24]
            private static byte[] aes192Decrypt24(byte[] input, byte[] key) {
                byte[] output = new byte[3 + 4 * (NB - 1) + 1];
                byte[][] state = new byte[4][NB];
                byte[][] w = new byte[4][NB * (NR + 1)];
                int r, c, round;

                for (r = 0; r <= 3; r++)
                    for (c = 0; c <= NB - 1; c++)
                        state[r][c] = input[r + 4 * c];

                w = KeyExpansion(key, w);
                state = addRoundKey(state, w, NR * NB);

                for (round = NR - 1; round >= 1; round--) {
                    state = invShiftRows(state);
                    state = invSubBytes(state);
                    state = addRoundKey(state, w, round * NB);
                    state = invMixColumns(state);
                }

                state = invShiftRows(state);
                state = invSubBytes(state);
                state = addRoundKey(state, w, 0);

                for (r = 0; r <= 3; r++)
                    for (c = 0; c <= NB - 1; c++)
                        output[r + 4 * c] = state[r][c];

                return output;
            }

            private static byte[] aes192EncryptCbcIvResult = null;
            private static byte[] aes192EncryptCbc(byte[] input, int inputLength, byte[] key, byte[] iv, byte[] mac_key) {
                byte[] curData = new byte[24];
                Arrays.fill(curData, (byte)0x00);
                byte[] curIV = new byte[24];
                Arrays.fill(curIV, (byte)0x00);
                int i;

                if (inputLength == 0)
                    return null;

                //записываем и дополняем данные при необходимости
                System.arraycopy(input, 0, curData, 0, (inputLength>24)?24:inputLength);

                System.arraycopy(iv, 0, curIV, 0, 24);

                byte[] output = new byte[24 * (int)Math.ceil(inputLength / 24)];
                i = 0;
                while (i < inputLength) {
                    curData = xor192(curData, curIV);

                    if ((mac_key != null)&&(i+24 == inputLength))
                        curData = xor192(curData, mac_key);

                    byte[] tmp = aes192Encrypt24(curData, key);
                    System.arraycopy(tmp, 0, output, i, tmp.length);

                    System.arraycopy(tmp, 0, curIV, 0, 24);

                    i += 24;
                    if (i < inputLength)
                        System.arraycopy(input, i, curData, 0, (inputLength>24+i)?24:(inputLength-i));
                }

                if (aes192EncryptCbcIvResult != null)
                    System.arraycopy(curIV, 0, aes192EncryptCbcIvResult, 0, 24);

                return output;
            }

            private static byte[] aes128EncryptCbcIvResult = new byte[16];
            private static byte[] aes128EncryptCbc(byte[] input, int inputLength, byte[] key, byte[] iv, byte[] mac_key) {
                byte[] curData = new byte[16];
                //Arrays.fill(curData, (byte)0x00);
                byte[] curIV = new byte[16];
                //Arrays.fill(curIV, (byte)0x00);
                int i;

                if (inputLength == 0)
                    return null;
                //записываем и дополняем данные при необходимости
                System.arraycopy(input, 0, curData, 0, (inputLength>16)?16:inputLength);

                System.arraycopy(iv, 0, curIV, 0, 16);

                byte[] output = new byte[16 * (int)Math.ceil(inputLength / 16)];
                i = 0;
                while (i < inputLength) {
                    curData = xor128(curData, curIV);

                    if ((mac_key != null)&&(i+16 == inputLength))
                        curData = xor128(curData, mac_key);

                    byte[] tmp = aes128Encrypt16(curData, key);
                    System.arraycopy(tmp, 0, output, i, tmp.length);

                    System.arraycopy(tmp, 0, curIV, 0, 16);

                    i += 16;
                    if (i < inputLength)
                        System.arraycopy(input, i, curData, 0, (inputLength > (16 + i)) ? 16 : (inputLength - i));
                }

                if (aes128EncryptCbcIvResult != null)
                    System.arraycopy(curIV, 0, aes128EncryptCbcIvResult, 0, 16);

                return output;
            }

            private static byte[] aes128DecryptCbcIvResult = new byte[16];
            private static byte[] aes128DecryptCbc(byte[] input, int inputLength, byte[] key, byte[] iv, byte[] mac_key) {
                byte[] curData = new byte[16];
                //Arrays.fill(curData, (byte)0x00);
                byte[] curIV = new byte[16];
                //Arrays.fill(curIV, (byte)0x00);
                int i = 0;

                if ((inputLength == 0) || ((inputLength % 16) != 0))
                    return null;

                System.arraycopy(iv, 0, curIV, 0, 16);

                byte[] output = new byte[16 * (int)Math.ceil(inputLength / 16)];

                while (i < inputLength) {
                    curData = aes128Decrypt16(Arrays.copyOfRange(input, i, input.length), key);

                    //System.arraycopy(tmp, 0, curData, 0, tmp.length);

                    curData = xor128(curData, curIV);

                    if (mac_key != null)
                        curData = xor128(curData, mac_key);

                    System.arraycopy(curData, 0, output, i, 16);

                    System.arraycopy(input, i, curIV, 0, 16);

                    i += 16;
                }

                if (aes128DecryptCbcIvResult != null)
                    System.arraycopy(input, i-16, aes128DecryptCbcIvResult, 0, 16);

                return output;
            }

            //расширение ключа - формирование ключей для всех раундов шифрования
            private static byte[][] KeyExpansion(byte[] key, byte[][] w) {
                int i, j;
                byte[] temp = new byte[4];

                for (i = 0; i <= NK - 1; i++)
                    for (j = 0; j <= 3; j++)
                        w[j][i] = key[4 * i + j];

                for (i = NK; i <= NB * (NR + 1) - 1; i++) {
                    for (j = 0; j <= 3; j++)
                        temp[j] = w[j][i - 1];
                    if (i % NK == 0) {
                        temp = rotWord(temp);
                        temp = subWord(temp);
                        temp[0] = (byte)(temp[0] ^ RCON[i / NK]);
                    } else if (NK > 6 && i % NK == 4)
                        temp = subWord(temp);
                    for (j = 0; j <= 3; j++)
                        w[j][i] = (byte)((byte)w[j][i - NK] ^ (byte)temp[j]);
                }

                return w;
            }

            private static byte[] xor128(byte[] a, byte[] b) {
                return xorSize(a, b,16);
            }

            private static byte[] xor192(byte[] a, byte[] b) {
                return xorSize(a, b,24);
            }

            private static byte[] xorSize(byte[] a, byte[] b) {
                return xorSize(a, b,16);
            }

            private static byte[] xorSize(byte[] a, byte[] b, int Size) {
                byte[] out = new byte[Size];
                int i;
                for (i=0;i<Size; i++)
                    out[i] = (byte)(a[i] ^ b[i]);
                return out;
            }

            private static byte[] leftshiftOnebit(byte[] input) {
                return leftshiftOnebit(input, 16);
            }

            private static byte[] leftshiftOnebit(byte[] input, int Size) {
                byte[] output = new byte[Size];
                int i;
                byte overflow = 0;

                for ( i=Size-1; i>=0; i-- ) {
                    output[i] = (byte)(input[i] << 1);
                    output[i] |= overflow;
                    overflow = (byte)(((byte)(input[i] & 0x80) == (byte)0x80)?1:0);
                }

                return output;
            }

            private static byte[][] generateSubkey(byte[] key) {
                byte[] const_Rb = new  byte[] {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, (byte)0x87};

                byte[] L = new byte[16];
                byte[] Z = new byte[16];
                byte[] tmp = new byte[16];
                int i;

                byte[] K1;
                byte[] K2;

                for ( i=0; i<16; i++ )
                    Z[i] = 0;

                L = aes128Encrypt16( Z, key );

                if ( (L[0] & 0x80) == 0 ) { /* If MSB(L) = 0, then K1 = L << 1 */
                    K1 = leftshiftOnebit(L);
                } else { /* Else K1 = ( L << 1 ) (+) Rb */

                    tmp = leftshiftOnebit(L);
                    K1 = xor128(tmp,const_Rb);
                }

                if ( (K1[0] & 0x80) == 0 ) {
                    K2 = leftshiftOnebit(K1);
                } else {
                    tmp = leftshiftOnebit(K1);
                    K2 = xor128(tmp,const_Rb);
                }

                return new byte[][]{K1, K2};
            }

            //табличная замена всех байт числа посредством S_BOX
            private static byte[] subWord(byte[] v) {
                int i;

                for (i = 0; i <= 3; i++)
                    v[i] = S_BOX[v[i] & 0xFF];

                return v;
            }

            //побайтный сдвиг числа: [a0,a1,a2,a3]->[a1,a2,a3,a0]
            private static byte[] rotWord(byte[] v) {
                byte temp = v[0];

                v[0] = v[1];
                v[1] = v[2];
                v[2] = v[3];
                v[3] = temp;

                return v;
            }

            //наложение на массив данных ключей шифровани w[start..start + NB - 1]
            private static byte[][] addRoundKey(byte[][] state, byte[][] w, int start) {
                int i, j;

                for (i = 0; i <= NB - 1; i++)
                    for (j = 0; j <= 3; j++)
                        state[j][i] ^= w[j][start + i];

                return state;
            }

            //табличная замена всех байт массива данных посредством S_BOX
            private static byte[][] subBytes(byte[][] state) {
                int i, j;

                for (i = 0; i <= NB - 1; i++)
                    for (j = 0; j <= 3; j++)
                        state[j][i] = S_BOX[state[j][i] & 0xFF];

                return state;
            }

            //циклический сдвиг строк массива данных (i-строка сдвигвается влево
            //на i шагов)
            private static byte[][] shiftRows(byte[][] state) {
                byte[] temp = new byte[NB];
                int i, j, n;

                for (i = 1; i <= 3; i++) {
                    for (j = 0; j <= NB - 1; j++)
                        temp[j] = state[i][j];
                    for (j = 0; j <= NB - 1; j++) {
                        n = (j - i < 0) ? (j - i + NB) : (j - i);
                        state[i][n] = temp[j];
                    }
                }

                return state;
            }

            //результат умножения полинома x на 2 в поле GF(2^8)
            private final static byte[] X_MUL_2 = new byte[] {
                    0x00, 0x02, 0x04, 0x06, 0x08, 0x0a, 0x0c, 0x0e, 0x10, 0x12, 0x14, 0x16, 0x18, 0x1a, 0x1c, 0x1e,
                    0x20, 0x22, 0x24, 0x26, 0x28, 0x2a, 0x2c, 0x2e, 0x30, 0x32, 0x34, 0x36, 0x38, 0x3a, 0x3c, 0x3e,
                    0x40, 0x42, 0x44, 0x46, 0x48, 0x4a, 0x4c, 0x4e, 0x50, 0x52, 0x54, 0x56, 0x58, 0x5a, 0x5c, 0x5e,
                    0x60, 0x62, 0x64, 0x66, 0x68, 0x6a, 0x6c, 0x6e, 0x70, 0x72, 0x74, 0x76, 0x78, 0x7a, 0x7c, 0x7e,
                    (byte)0x80, (byte)0x82, (byte)0x84, (byte)0x86, (byte)0x88, (byte)0x8a, (byte)0x8c, (byte)0x8e, (byte)0x90, (byte)0x92, (byte)0x94, (byte)0x96, (byte)0x98, (byte)0x9a, (byte)0x9c, (byte)0x9e,
                    (byte)0xa0, (byte)0xa2, (byte)0xa4, (byte)0xa6, (byte)0xa8, (byte)0xaa, (byte)0xac, (byte)0xae, (byte)0xb0, (byte)0xb2, (byte)0xb4, (byte)0xb6, (byte)0xb8, (byte)0xba, (byte)0xbc, (byte)0xbe,
                    (byte)0xc0, (byte)0xc2, (byte)0xc4, (byte)0xc6, (byte)0xc8, (byte)0xca, (byte)0xcc, (byte)0xce, (byte)0xd0, (byte)0xd2, (byte)0xd4, (byte)0xd6, (byte)0xd8, (byte)0xda, (byte)0xdc, (byte)0xde,
                    (byte)0xe0, (byte)0xe2, (byte)0xe4, (byte)0xe6, (byte)0xe8, (byte)0xea, (byte)0xec, (byte)0xee, (byte)0xf0, (byte)0xf2, (byte)0xf4, (byte)0xf6, (byte)0xf8, (byte)0xfa, (byte)0xfc, (byte)0xfe,
                    0x1b, 0x19, 0x1f, 0x1d, 0x13, 0x11, 0x17, 0x15, 0x0b, 0x09, 0x0f, 0x0d, 0x03, 0x01, 0x07, 0x05,
                    0x3b, 0x39, 0x3f, 0x3d, 0x33, 0x31, 0x37, 0x35, 0x2b, 0x29, 0x2f, 0x2d, 0x23, 0x21, 0x27, 0x25,
                    0x5b, 0x59, 0x5f, 0x5d, 0x53, 0x51, 0x57, 0x55, 0x4b, 0x49, 0x4f, 0x4d, 0x43, 0x41, 0x47, 0x45,
                    0x7b, 0x79, 0x7f, 0x7d, 0x73, 0x71, 0x77, 0x75, 0x6b, 0x69, 0x6f, 0x6d, 0x63, 0x61, 0x67, 0x65,
                    (byte)0x9b, (byte)0x99, (byte)0x9f, (byte)0x9d, (byte)0x93, (byte)0x91, (byte)0x97, (byte)0x95, (byte)0x8b, (byte)0x89, (byte)0x8f, (byte)0x8d, (byte)0x83, (byte)0x81, (byte)0x87, (byte)0x85,
                    (byte)0xbb, (byte)0xb9, (byte)0xbf, (byte)0xbd, (byte)0xb3, (byte)0xb1, (byte)0xb7, (byte)0xb5, (byte)0xab, (byte)0xa9, (byte)0xaf, (byte)0xad, (byte)0xa3, (byte)0xa1, (byte)0xa7, (byte)0xa5,
                    (byte)0xdb, (byte)0xd9, (byte)0xdf, (byte)0xdd, (byte)0xd3, (byte)0xd1, (byte)0xd7, (byte)0xd5, (byte)0xcb, (byte)0xc9, (byte)0xcf, (byte)0xcd, (byte)0xc3, (byte)0xc1, (byte)0xc7, (byte)0xc5,
                    (byte)0xfb, (byte)0xf9, (byte)0xff, (byte)0xfd, (byte)0xf3, (byte)0xf1, (byte)0xf7, (byte)0xf5, (byte)0xeb, (byte)0xe9, (byte)0xef, (byte)0xed, (byte)0xe3, (byte)0xe1, (byte)0xe7, (byte)0xe5};

            //результат умножения полинома x на 3 в поле GF(2^8)
            private final static byte[] X_MUL_3 = new byte[] {
                    0x00, 0x03, 0x06, 0x05, 0x0c, 0x0f, 0x0a, 0x09, 0x18, 0x1b, 0x1e, 0x1d, 0x14, 0x17, 0x12, 0x11,
                    0x30, 0x33, 0x36, 0x35, 0x3c, 0x3f, 0x3a, 0x39, 0x28, 0x2b, 0x2e, 0x2d, 0x24, 0x27, 0x22, 0x21,
                    0x60, 0x63, 0x66, 0x65, 0x6c, 0x6f, 0x6a, 0x69, 0x78, 0x7b, 0x7e, 0x7d, 0x74, 0x77, 0x72, 0x71,
                    0x50, 0x53, 0x56, 0x55, 0x5c, 0x5f, 0x5a, 0x59, 0x48, 0x4b, 0x4e, 0x4d, 0x44, 0x47, 0x42, 0x41,
                    (byte)0xc0, (byte)0xc3, (byte)0xc6, (byte)0xc5, (byte)0xcc, (byte)0xcf, (byte)0xca, (byte)0xc9, (byte)0xd8, (byte)0xdb, (byte)0xde, (byte)0xdd, (byte)0xd4, (byte)0xd7, (byte)0xd2, (byte)0xd1,
                    (byte)0xf0, (byte)0xf3, (byte)0xf6, (byte)0xf5, (byte)0xfc, (byte)0xff, (byte)0xfa, (byte)0xf9, (byte)0xe8, (byte)0xeb, (byte)0xee, (byte)0xed, (byte)0xe4, (byte)0xe7, (byte)0xe2, (byte)0xe1,
                    (byte)0xa0, (byte)0xa3, (byte)0xa6, (byte)0xa5, (byte)0xac, (byte)0xaf, (byte)0xaa, (byte)0xa9, (byte)0xb8, (byte)0xbb, (byte)0xbe, (byte)0xbd, (byte)0xb4, (byte)0xb7, (byte)0xb2, (byte)0xb1,
                    (byte)0x90, (byte)0x93, (byte)0x96, (byte)0x95, (byte)0x9c, (byte)0x9f, (byte)0x9a, (byte)0x99, (byte)0x88, (byte)0x8b, (byte)0x8e, (byte)0x8d, (byte)0x84, (byte)0x87, (byte)0x82, (byte)0x81,
                    (byte)0x9b, (byte)0x98, (byte)0x9d, (byte)0x9e, (byte)0x97, (byte)0x94, (byte)0x91, (byte)0x92, (byte)0x83, (byte)0x80, (byte)0x85, (byte)0x86, (byte)0x8f, (byte)0x8c, (byte)0x89, (byte)0x8a,
                    (byte)0xab, (byte)0xa8, (byte)0xad, (byte)0xae, (byte)0xa7, (byte)0xa4, (byte)0xa1, (byte)0xa2, (byte)0xb3, (byte)0xb0, (byte)0xb5, (byte)0xb6, (byte)0xbf, (byte)0xbc, (byte)0xb9, (byte)0xba,
                    (byte)0xfb, (byte)0xf8, (byte)0xfd, (byte)0xfe, (byte)0xf7, (byte)0xf4, (byte)0xf1, (byte)0xf2, (byte)0xe3, (byte)0xe0, (byte)0xe5, (byte)0xe6, (byte)0xef, (byte)0xec, (byte)0xe9, (byte)0xea,
                    (byte)0xcb, (byte)0xc8, (byte)0xcd, (byte)0xce, (byte)0xc7, (byte)0xc4, (byte)0xc1, (byte)0xc2, (byte)0xd3, (byte)0xd0, (byte)0xd5, (byte)0xd6, (byte)0xdf, (byte)0xdc, (byte)0xd9, (byte)0xda,
                    0x5b, 0x58, 0x5d, 0x5e, 0x57, 0x54, 0x51, 0x52, 0x43, 0x40, 0x45, 0x46, 0x4f, 0x4c, 0x49, 0x4a,
                    0x6b, 0x68, 0x6d, 0x6e, 0x67, 0x64, 0x61, 0x62, 0x73, 0x70, 0x75, 0x76, 0x7f, 0x7c, 0x79, 0x7a,
                    0x3b, 0x38, 0x3d, 0x3e, 0x37, 0x34, 0x31, 0x32, 0x23, 0x20, 0x25, 0x26, 0x2f, 0x2c, 0x29, 0x2a,
                    0x0b, 0x08, 0x0d, 0x0e, 0x07, 0x04, 0x01, 0x02, 0x13, 0x10, 0x15, 0x16, 0x1f, 0x1c, 0x19, 0x1a};

            //смешение столбцов
            private static byte[][] mixColumns(byte[][] state) {
                byte[] temp = new byte[4];
                int i, j;

                for (i = 0; i <= NB - 1; i++) {
                    for (j = 0; j <= 3; j++)
                        temp[j] = state[j][i];
                    state[0][i] = (byte)(X_MUL_2[temp[0] & 0xFF] ^ X_MUL_3[temp[1] & 0xFF] ^         temp[2]  ^         temp[3]);
                    state[1][i] = (byte)(        temp[0]  ^ X_MUL_2[temp[1] & 0xFF] ^ X_MUL_3[temp[2] & 0xFF] ^         temp[3]);
                    state[2][i] = (byte)(        temp[0]  ^         temp[1]  ^ X_MUL_2[temp[2] & 0xFF] ^ X_MUL_3[temp[3] & 0xFF]);
                    state[3][i] = (byte)(X_MUL_3[temp[0] & 0xFF] ^         temp[1]  ^         temp[2]  ^ X_MUL_2[temp[3] & 0xFF]);
                }

                return state;
            }

            //массив, определяющий нелинейную подстановку между двумя байтами
            private final static byte[] S_BOX = new byte[] {
                    0x63, 0x7c, 0x77, 0x7b, (byte)0xf2, 0x6b, 0x6f, (byte)0xc5, 0x30, 0x01, 0x67, 0x2b, (byte)0xfe, (byte)0xd7, (byte)0xab, 0x76,
                    (byte)0xca, (byte)0x82, (byte)0xc9, 0x7d, (byte)0xfa, 0x59, 0x47, (byte)0xf0, (byte)0xad, (byte)0xd4, (byte)0xa2, (byte)0xaf, (byte)0x9c, (byte)0xa4, 0x72, (byte)0xc0,
                    (byte)0xb7, (byte)0xfd, (byte)0x93, 0x26, 0x36, 0x3f, (byte)0xf7, (byte)0xcc, 0x34, (byte)0xa5, (byte)0xe5, (byte)0xf1, 0x71, (byte)0xd8, 0x31, 0x15,
                    0x04, (byte)0xc7, 0x23, (byte)0xc3, 0x18, (byte)0x96, 0x05, (byte)0x9a, 0x07, 0x12, (byte)0x80, (byte)0xe2, (byte)0xeb, 0x27, (byte)0xb2, 0x75,
                    0x09, (byte)0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, (byte)0xa0, 0x52, 0x3b, (byte)0xd6, (byte)0xb3, 0x29, (byte)0xe3, 0x2f, (byte)0x84,
                    0x53, (byte)0xd1, 0x00, (byte)0xed, 0x20, (byte)0xfc, (byte)0xb1, 0x5b, 0x6a, (byte)0xcb, (byte)0xbe, 0x39, 0x4a, 0x4c, 0x58, (byte)0xcf,
                    (byte)0xd0, (byte)0xef, (byte)0xaa, (byte)0xfb, 0x43, 0x4d, 0x33, (byte)0x85, 0x45, (byte)0xf9, 0x02, 0x7f, 0x50, 0x3c, (byte)0x9f, (byte)0xa8,
                    0x51, (byte)0xa3, 0x40, (byte)0x8f, (byte)0x92, (byte)0x9d, 0x38, (byte)0xf5, (byte)0xbc, (byte)0xb6, (byte)0xda, 0x21, 0x10, (byte)0xff, (byte)0xf3, (byte)0xd2,
                    (byte)0xcd, 0x0c, 0x13, (byte)0xec, 0x5f, (byte)0x97, 0x44, 0x17, (byte)0xc4, (byte)0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
                    0x60, (byte)0x81, 0x4f, (byte)0xdc, 0x22, 0x2a, (byte)0x90, (byte)0x88, 0x46, (byte)0xee, (byte)0xb8, 0x14, (byte)0xde, 0x5e, 0x0b, (byte)0xdb,
                    (byte)0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, (byte)0xc2, (byte)0xd3, (byte)0xac, 0x62, (byte)0x91, (byte)0x95, (byte)0xe4, 0x79,
                    (byte)0xe7, (byte)0xc8, 0x37, 0x6d, (byte)0x8d, (byte)0xd5, 0x4e, (byte)0xa9, 0x6c, 0x56, (byte)0xf4, (byte)0xea, 0x65, 0x7a, (byte)0xae, 0x08,
                    (byte)0xba, 0x78, 0x25, 0x2e, 0x1c, (byte)0xa6, (byte)0xb4, (byte)0xc6, (byte)0xe8, (byte)0xdd, 0x74, 0x1f, 0x4b, (byte)0xbd, (byte)0x8b, (byte)0x8a,
                    0x70, 0x3e, (byte)0xb5, 0x66, 0x48, 0x03, (byte)0xf6, 0x0e, 0x61, 0x35, 0x57, (byte)0xb9, (byte)0x86, (byte)0xc1, 0x1d, (byte)0x9e,
                    (byte)0xe1, (byte)0xf8, (byte)0x98, 0x11, 0x69, (byte)0xd9, (byte)0x8e, (byte)0x94, (byte)0x9b, 0x1e, (byte)0x87, (byte)0xe9, (byte)0xce, 0x55, 0x28, (byte)0xdf,
                    (byte)0x8c, (byte)0xa1, (byte)0x89, 0x0d, (byte)0xbf, (byte)0xe6, 0x42, 0x68, 0x41, (byte)0x99, 0x2d, 0x0f, (byte)0xb0, 0x54, (byte)0xbb, 0x16};

            //константы для процедуры расширения ключа
            private final static byte[] RCON = new byte[]{0x00, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, (byte)0x80, 0x1b, 0x36};

            //циклический сдвиг строк массива данных (i-строка сдвигвается вправо
            //на i шагов)
            private static byte[][] invShiftRows(byte[][] state) {
                byte[] temp = new byte[NB];
                int i, j, n;

                for (i = 1; i <= 3; i++) {
                    for (j = 0; j <= NB - 1; j++)
                        temp[j] = state[i][j];
                    for (j = 0; j <= NB - 1; j++) {
                        n = (j + i > NB - 1) ? (j + i - NB) : (j + i);
                        state[i][n] = temp[j];
                    }
                }

                return state;
            }

            //массив, определяющий нелинейную подстановку между двумя байтами
            private final static byte[] INV_S_BOX = new byte[] {
                    0x52, 0x09, 0x6a, (byte)0xd5, 0x30, 0x36, (byte)0xa5, 0x38, (byte)0xbf, 0x40, (byte)0xa3, (byte)0x9e, (byte)0x81, (byte)0xf3, (byte)0xd7, (byte)0xfb,
                    0x7c, (byte)0xe3, 0x39, (byte)0x82, (byte)0x9b, 0x2f, (byte)0xff, (byte)0x87, 0x34, (byte)0x8e, 0x43, 0x44, (byte)0xc4, (byte)0xde, (byte)0xe9, (byte)0xcb,
                    0x54, 0x7b, (byte)0x94, 0x32, (byte)0xa6, (byte)0xc2, 0x23, 0x3d, (byte)0xee, 0x4c, (byte)0x95, 0x0b, 0x42, (byte)0xfa, (byte)0xc3, 0x4e,
                    0x08, 0x2e, (byte)0xa1, 0x66, 0x28, (byte)0xd9, 0x24, (byte)0xb2, 0x76, 0x5b, (byte)0xa2, 0x49, 0x6d, (byte)0x8b, (byte)0xd1, 0x25,
                    0x72, (byte)0xf8, (byte)0xf6, 0x64, (byte)0x86, 0x68, (byte)0x98, 0x16, (byte)0xd4, (byte)0xa4, 0x5c, (byte)0xcc, 0x5d, 0x65, (byte)0xb6, (byte)0x92,
                    0x6c, 0x70, 0x48, 0x50, (byte)0xfd, (byte)0xed, (byte)0xb9, (byte)0xda, 0x5e, 0x15, 0x46, 0x57, (byte)0xa7, (byte)0x8d, (byte)0x9d, (byte)0x84,
                    (byte)0x90, (byte)0xd8, (byte)0xab, 0x00, (byte)0x8c, (byte)0xbc, (byte)0xd3, 0x0a, (byte)0xf7, (byte)0xe4, 0x58, 0x05, (byte)0xb8, (byte)0xb3, 0x45, 0x06,
                    (byte)0xd0, 0x2c, 0x1e, (byte)0x8f, (byte)0xca, 0x3f, 0x0f, 0x02, (byte)0xc1, (byte)0xaf, (byte)0xbd, 0x03, 0x01, 0x13, (byte)0x8a, 0x6b,
                    0x3a, (byte)0x91, 0x11, 0x41, 0x4f, 0x67, (byte)0xdc, (byte)0xea, (byte)0x97, (byte)0xf2, (byte)0xcf, (byte)0xce, (byte)0xf0, (byte)0xb4, (byte)0xe6, 0x73,
                    (byte)0x96, (byte)0xac, 0x74, 0x22, (byte)0xe7, (byte)0xad, 0x35, (byte)0x85, (byte)0xe2, (byte)0xf9, 0x37, (byte)0xe8, 0x1c, 0x75, (byte)0xdf, 0x6e,
                    0x47, (byte)0xf1, 0x1a, 0x71, 0x1d, 0x29, (byte)0xc5, (byte)0x89, 0x6f, (byte)0xb7, 0x62, 0x0e, (byte)0xaa, 0x18, (byte)0xbe, 0x1b,
                    (byte)0xfc, 0x56, 0x3e, 0x4b, (byte)0xc6, (byte)0xd2, 0x79, 0x20, (byte)0x9a, (byte)0xdb, (byte)0xc0, (byte)0xfe, 0x78, (byte)0xcd, 0x5a, (byte)0xf4,
                    0x1f, (byte)0xdd, (byte)0xa8, 0x33, (byte)0x88, 0x07, (byte)0xc7, 0x31, (byte)0xb1, 0x12, 0x10, 0x59, 0x27, (byte)0x80, (byte)0xec, 0x5f,
                    0x60, 0x51, 0x7f, (byte)0xa9, 0x19, (byte)0xb5, 0x4a, 0x0d, 0x2d, (byte)0xe5, 0x7a, (byte)0x9f, (byte)0x93, (byte)0xc9, (byte)0x9c, (byte)0xef,
                    (byte)0xa0, (byte)0xe0, 0x3b, 0x4d, (byte)0xae, 0x2a, (byte)0xf5, (byte)0xb0, (byte)0xc8, (byte)0xeb, (byte)0xbb, 0x3c, (byte)0x83, 0x53, (byte)0x99, 0x61,
                    0x17, 0x2b, 0x04, 0x7e, (byte)0xba, 0x77, (byte)0xd6, 0x26, (byte)0xe1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0c, 0x7d};

            //табличная замена всех байт массива данных посредством INV_S_BOX
            private static byte[][] invSubBytes(byte[][] state) {
                int i, j;

                for (i = 0; i <= NB - 1; i++)
                    for (j = 0; j <= 3; j++)
                        state[j][i] = INV_S_BOX[state[j][i] & 0xFF];

                return state;
            }

            //результат умножения полинома x на 0x0E в поле GF(2^8)
            private final static byte[] X_MUL_E = new byte[] {
                    0x00, 0x0e, 0x1c, 0x12, 0x38, 0x36, 0x24, 0x2a, 0x70, 0x7e, 0x6c, 0x62, 0x48, 0x46, 0x54, 0x5a,
                    (byte)0xe0, (byte)0xee, (byte)0xfc, (byte)0xf2, (byte)0xd8, (byte)0xd6, (byte)0xc4, (byte)0xca, (byte)0x90, (byte)0x9e, (byte)0x8c, (byte)0x82, (byte)0xa8, (byte)0xa6, (byte)0xb4, (byte)0xba,
                    (byte)0xdb, (byte)0xd5, (byte)0xc7, (byte)0xc9, (byte)0xe3, (byte)0xed, (byte)0xff, (byte)0xf1, (byte)0xab, (byte)0xa5, (byte)0xb7, (byte)0xb9, (byte)0x93, (byte)0x9d, (byte)0x8f, (byte)0x81,
                    0x3b, 0x35, 0x27, 0x29, 0x03, 0x0d, 0x1f, 0x11, 0x4b, 0x45, 0x57, 0x59, 0x73, 0x7d, 0x6f, 0x61,
                    (byte)0xad, (byte)0xa3, (byte)0xb1, (byte)0xbf, (byte)0x95, (byte)0x9b, (byte)0x89, (byte)0x87, (byte)0xdd, (byte)0xd3, (byte)0xc1, (byte)0xcf, (byte)0xe5, (byte)0xeb, (byte)0xf9, (byte)0xf7,
                    0x4d, 0x43, 0x51, 0x5f, 0x75, 0x7b, 0x69, 0x67, 0x3d, 0x33, 0x21, 0x2f, 0x05, 0x0b, 0x19, 0x17,
                    0x76, 0x78, 0x6a, 0x64, 0x4e, 0x40, 0x52, 0x5c, 0x06, 0x08, 0x1a, 0x14, 0x3e, 0x30, 0x22, 0x2c,
                    (byte)0x96, (byte)0x98, (byte)0x8a, (byte)0x84, (byte)0xae, (byte)0xa0, (byte)0xb2, (byte)0xbc, (byte)0xe6, (byte)0xe8, (byte)0xfa, (byte)0xf4, (byte)0xde, (byte)0xd0, (byte)0xc2, (byte)0xcc,
                    0x41, 0x4f, 0x5d, 0x53, 0x79, 0x77, 0x65, 0x6b, 0x31, 0x3f, 0x2d, 0x23, 0x09, 0x07, 0x15, 0x1b,
                    (byte)0xa1, (byte)0xaf, (byte)0xbd, (byte)0xb3, (byte)0x99, (byte)0x97, (byte)0x85, (byte)0x8b, (byte)0xd1, (byte)0xdf, (byte)0xcd, (byte)0xc3, (byte)0xe9, (byte)0xe7, (byte)0xf5, (byte)0xfb,
                    (byte)0x9a, (byte)0x94, (byte)0x86, (byte)0x88, (byte)0xa2, (byte)0xac, (byte)0xbe, (byte)0xb0, (byte)0xea, (byte)0xe4, (byte)0xf6, (byte)0xf8, (byte)0xd2, (byte)0xdc, (byte)0xce, (byte)0xc0,
                    0x7a, 0x74, 0x66, 0x68, 0x42, 0x4c, 0x5e, 0x50, 0x0a, 0x04, 0x16, 0x18, 0x32, 0x3c, 0x2e, 0x20,
                    (byte)0xec, (byte)0xe2, (byte)0xf0, (byte)0xfe, (byte)0xd4, (byte)0xda, (byte)0xc8, (byte)0xc6, (byte)0x9c, (byte)0x92, (byte)0x80, (byte)0x8e, (byte)0xa4, (byte)0xaa, (byte)0xb8, (byte)0xb6,
                    0x0c, 0x02, 0x10, 0x1e, 0x34, 0x3a, 0x28, 0x26, 0x7c, 0x72, 0x60, 0x6e, 0x44, 0x4a, 0x58, 0x56,
                    0x37, 0x39, 0x2b, 0x25, 0x0f, 0x01, 0x13, 0x1d, 0x47, 0x49, 0x5b, 0x55, 0x7f, 0x71, 0x63, 0x6d,
                    (byte)0xd7, (byte)0xd9, (byte)0xcb, (byte)0xc5, (byte)0xef, (byte)0xe1, (byte)0xf3, (byte)0xfd, (byte)0xa7, (byte)0xa9, (byte)0xbb, (byte)0xb5, (byte)0x9f, (byte)0x91, (byte)0x83, (byte)0x8d};

            //результат умножения полинома x на 0x0B в поле GF(2^8)
            private final static byte[] X_MUL_B = new byte[] {
                    0x00, 0x0b, 0x16, 0x1d, 0x2c, 0x27, 0x3a, 0x31, 0x58, 0x53, 0x4e, 0x45, 0x74, 0x7f, 0x62, 0x69,
                    (byte)0xb0, (byte)0xbb, (byte)0xa6, (byte)0xad, (byte)0x9c, (byte)0x97, (byte)0x8a, (byte)0x81, (byte)0xe8, (byte)0xe3, (byte)0xfe, (byte)0xf5, (byte)0xc4, (byte)0xcf, (byte)0xd2, (byte)0xd9,
                    0x7b, 0x70, 0x6d, 0x66, 0x57, 0x5c, 0x41, 0x4a, 0x23, 0x28, 0x35, 0x3e, 0x0f, 0x04, 0x19, 0x12,
                    (byte)0xcb, (byte)0xc0, (byte)0xdd, (byte)0xd6, (byte)0xe7, (byte)0xec, (byte)0xf1, (byte)0xfa, (byte)0x93,(byte) 0x98, (byte)0x85, (byte)0x8e, (byte)0xbf, (byte)0xb4, (byte)0xa9, (byte)0xa2,
                    (byte)0xf6, (byte)0xfd, (byte)0xe0, (byte)0xeb, (byte)0xda, (byte)0xd1, (byte)0xcc, (byte)0xc7, (byte)0xae, (byte)0xa5, (byte)0xb8, (byte)0xb3, (byte)0x82, (byte)0x89, (byte)0x94, (byte)0x9f,
                    0x46, 0x4d, 0x50, 0x5b, 0x6a, 0x61, 0x7c, 0x77, 0x1e, 0x15, 0x08, 0x03, 0x32, 0x39, 0x24, 0x2f,
                    (byte)0x8d, (byte)0x86, (byte)0x9b, (byte)0x90, (byte)0xa1, (byte)0xaa, (byte)0xb7, (byte)0xbc, (byte)0xd5, (byte)0xde, (byte)0xc3, (byte)0xc8, (byte)0xf9, (byte)0xf2, (byte)0xef, (byte)0xe4,
                    0x3d, 0x36, 0x2b, 0x20, 0x11, 0x1a, 0x07, 0x0c, 0x65, 0x6e, 0x73, 0x78, 0x49, 0x42, 0x5f, 0x54,
                    (byte)0xf7, (byte)0xfc, (byte)0xe1, (byte)0xea, (byte)0xdb, (byte)0xd0, (byte)0xcd, (byte)0xc6, (byte)0xaf, (byte)0xa4, (byte)0xb9, (byte)0xb2, (byte)0x83, (byte)0x88, (byte)0x95, (byte)0x9e,
                    0x47, 0x4c, 0x51, 0x5a, 0x6b, 0x60, 0x7d, 0x76, 0x1f, 0x14, 0x09, 0x02, 0x33, 0x38, 0x25, 0x2e,
                    (byte)0x8c, (byte)0x87, (byte)0x9a, (byte)0x91, (byte)0xa0, (byte)0xab, (byte)0xb6, (byte)0xbd, (byte)0xd4, (byte)0xdf, (byte)0xc2, (byte)0xc9, (byte)0xf8, (byte)0xf3, (byte)0xee, (byte)0xe5,
                    0x3c, 0x37, 0x2a, 0x21, 0x10, 0x1b, 0x06, 0x0d, 0x64, 0x6f, 0x72, 0x79, 0x48, 0x43, 0x5e, 0x55,
                    0x01, 0x0a, 0x17, 0x1c, 0x2d, 0x26, 0x3b, 0x30, 0x59, 0x52, 0x4f, 0x44, 0x75, 0x7e, 0x63, 0x68,
                    (byte)0xb1, (byte)0xba, (byte)0xa7, (byte)0xac, (byte)0x9d, (byte)0x96, (byte)0x8b, (byte)0x80, (byte)0xe9, (byte)0xe2, (byte)0xff, (byte)0xf4, (byte)0xc5, (byte)0xce, (byte)0xd3, (byte)0xd8,
                    0x7a, 0x71, 0x6c, 0x67, 0x56, 0x5d, 0x40, 0x4b, 0x22, 0x29, 0x34, 0x3f, 0x0e, 0x05, 0x18, 0x13,
                    (byte)0xca, (byte)0xc1, (byte)0xdc, (byte)0xd7, (byte)0xe6, (byte)0xed, (byte)0xf0, (byte)0xfb, (byte)0x92, (byte)0x99, (byte)0x84, (byte)0x8f, (byte)0xbe, (byte)0xb5, (byte)0xa8, (byte)0xa3};

            //результат умножения полинома x на 0x0D в поле GF(2^8)
            private final static byte[] X_MUL_D = new byte[] {
                    0x00, 0x0d, 0x1a, 0x17, 0x34, 0x39, 0x2e, 0x23, 0x68, 0x65, 0x72, 0x7f, 0x5c, 0x51, 0x46, 0x4b,
                    (byte)0xd0, (byte)0xdd, (byte)0xca, (byte)0xc7, (byte)0xe4, (byte)0xe9, (byte)0xfe, (byte)0xf3, (byte)0xb8, (byte)0xb5, (byte)0xa2, (byte)0xaf, (byte)0x8c, (byte)0x81, (byte)0x96, (byte)0x9b,
                    (byte)0xbb, (byte)0xb6, (byte)0xa1, (byte)0xac, (byte)0x8f, (byte)0x82, (byte)0x95, (byte)0x98, (byte)0xd3, (byte)0xde, (byte)0xc9, (byte)0xc4, (byte)0xe7, (byte)0xea, (byte)0xfd, (byte)0xf0,
                    0x6b, 0x66, 0x71, 0x7c, 0x5f, 0x52, 0x45, 0x48, 0x03, 0x0e, 0x19, 0x14, 0x37, 0x3a, 0x2d, 0x20,
                    0x6d, 0x60, 0x77, 0x7a, 0x59, 0x54, 0x43, 0x4e, 0x05, 0x08, 0x1f, 0x12, 0x31, 0x3c, 0x2b, 0x26,
                    (byte)0xbd, (byte)0xb0, (byte)0xa7, (byte)0xaa, (byte)0x89, (byte)0x84, (byte)0x93, (byte)0x9e, (byte)0xd5, (byte)0xd8, (byte)0xcf, (byte)0xc2, (byte)0xe1, (byte)0xec, (byte)0xfb, (byte)0xf6,
                    (byte)0xd6, (byte)0xdb, (byte)0xcc, (byte)0xc1, (byte)0xe2, (byte)0xef, (byte)0xf8, (byte)0xf5, (byte)0xbe, (byte)0xb3, (byte)0xa4, (byte)0xa9, (byte)0x8a, (byte)0x87, (byte)0x90, (byte)0x9d,
                    0x06, 0x0b, 0x1c, 0x11, 0x32, 0x3f, 0x28, 0x25, 0x6e, 0x63, 0x74, 0x79, 0x5a, 0x57, 0x40, 0x4d,
                    (byte)0xda, (byte)0xd7, (byte)0xc0, (byte)0xcd, (byte)0xee, (byte)0xe3, (byte)0xf4, (byte)0xf9, (byte)0xb2, (byte)0xbf, (byte)0xa8, (byte)0xa5, (byte)0x86, (byte)0x8b, (byte)0x9c, (byte)0x91,
                    0x0a, 0x07, 0x10, 0x1d, 0x3e, 0x33, 0x24, 0x29, 0x62, 0x6f, 0x78, 0x75, 0x56, 0x5b, 0x4c, 0x41,
                    0x61, 0x6c, 0x7b, 0x76, 0x55, 0x58, 0x4f, 0x42, 0x09, 0x04, 0x13, 0x1e, 0x3d, 0x30, 0x27, 0x2a,
                    (byte)0xb1, (byte)0xbc, (byte)0xab, (byte)0xa6, (byte)0x85, (byte)0x88, (byte)0x9f, (byte)0x92, (byte)0xd9, (byte)0xd4, (byte)0xc3, (byte)0xce, (byte)0xed, (byte)0xe0, (byte)0xf7, (byte)0xfa,
                    (byte)0xb7, (byte)0xba, (byte)0xad, (byte)0xa0, (byte)0x83, (byte)0x8e, (byte)0x99, (byte)0x94, (byte)0xdf, (byte)0xd2, (byte)0xc5, (byte)0xc8, (byte)0xeb, (byte)0xe6, (byte)0xf1, (byte)0xfc,
                    0x67, 0x6a, 0x7d, 0x70, 0x53, 0x5e, 0x49, 0x44, 0x0f, 0x02, 0x15, 0x18, 0x3b, 0x36, 0x21, 0x2c,
                    0x0c, 0x01, 0x16, 0x1b, 0x38, 0x35, 0x22, 0x2f, 0x64, 0x69, 0x7e, 0x73, 0x50, 0x5d, 0x4a, 0x47,
                    (byte)0xdc, (byte)0xd1, (byte)0xc6, (byte)0xcb, (byte)0xe8, (byte)0xe5, (byte)0xf2, (byte)0xff, (byte)0xb4, (byte)0xb9, (byte)0xae, (byte)0xa3, (byte)0x80, (byte)0x8d, (byte)0x9a, (byte)0x97};

            //результат умножения полинома x на 0x09 в поле GF(2^8)
            private final static byte[] X_MUL_9 = new byte[] {
                    0x00, 0x09, 0x12, 0x1b, 0x24, 0x2d, 0x36, 0x3f, 0x48, 0x41, 0x5a, 0x53, 0x6c, 0x65, 0x7e, 0x77,
                    (byte)0x90, (byte)0x99, (byte)0x82, (byte)0x8b, (byte)0xb4, (byte)0xbd, (byte)0xa6, (byte)0xaf, (byte)0xd8, (byte)0xd1, (byte)0xca, (byte)0xc3, (byte)0xfc, (byte)0xf5, (byte)0xee, (byte)0xe7,
                    0x3b, 0x32, 0x29, 0x20, 0x1f, 0x16, 0x0d, 0x04, 0x73, 0x7a, 0x61, 0x68, 0x57, 0x5e, 0x45, 0x4c,
                    (byte)0xab, (byte)0xa2, (byte)0xb9, (byte)0xb0, (byte)0x8f, (byte)0x86, (byte)0x9d, (byte)0x94, (byte)0xe3, (byte)0xea, (byte)0xf1, (byte)0xf8, (byte)0xc7, (byte)0xce, (byte)0xd5, (byte)0xdc,
                    0x76, 0x7f, 0x64, 0x6d, 0x52, 0x5b, 0x40, 0x49, 0x3e, 0x37, 0x2c, 0x25, 0x1a, 0x13, 0x08, 0x01,
                    (byte)0xe6, (byte)0xef, (byte)0xf4, (byte)0xfd, (byte)0xc2, (byte)0xcb, (byte)0xd0, (byte)0xd9, (byte)0xae, (byte)0xa7, (byte)0xbc, (byte)0xb5, (byte)0x8a, (byte)0x83, (byte)0x98, (byte)0x91,
                    0x4d, 0x44, 0x5f, 0x56, 0x69, 0x60, 0x7b, 0x72, 0x05, 0x0c, 0x17, 0x1e, 0x21, 0x28, 0x33, 0x3a,
                    (byte)0xdd, (byte)0xd4, (byte)0xcf, (byte)0xc6, (byte)0xf9, (byte)0xf0, (byte)0xeb, (byte)0xe2, (byte)0x95, (byte)0x9c, (byte)0x87, (byte)(byte)0x8e, (byte)0xb1, (byte)0xb8, (byte)0xa3, (byte)0xaa,
                    (byte)0xec, (byte)0xe5, (byte)0xfe, (byte)0xf7, (byte)0xc8, (byte)0xc1, (byte)0xda, (byte)0xd3, (byte)0xa4, (byte)0xad, (byte)0xb6, (byte)0xbf, (byte)0x80, (byte)0x89, (byte)0x92, (byte)0x9b,
                    0x7c, 0x75, 0x6e, 0x67, 0x58, 0x51, 0x4a, 0x43, 0x34, 0x3d, 0x26, 0x2f, 0x10, 0x19, 0x02, 0x0b,
                    (byte)0xd7, (byte)0xde, (byte)0xc5, (byte)0xcc, (byte)0xf3, (byte)0xfa, (byte)0xe1, (byte)0xe8, (byte)0x9f, (byte)0x96, (byte)0x8d, (byte)0x84, (byte)0xbb, (byte)0xb2, (byte)0xa9, (byte)0xa0,
                    0x47, 0x4e, 0x55, 0x5c, 0x63, 0x6a, 0x71, 0x78, 0x0f, 0x06, 0x1d, 0x14, 0x2b, 0x22, 0x39, 0x30,
                    (byte)0x9a, (byte)0x93, (byte)0x88, (byte)0x81, (byte)0xbe, (byte)0xb7, (byte)0xac, (byte)0xa5, (byte)0xd2, (byte)0xdb, (byte)0xc0, (byte)0xc9, (byte)0xf6, (byte)0xff, (byte)0xe4, (byte)0xed,
                    0x0a, 0x03, 0x18, 0x11, 0x2e, 0x27, 0x3c, 0x35, 0x42, 0x4b, 0x50, 0x59, 0x66, 0x6f, 0x74, 0x7d,
                    (byte)0xa1, (byte)0xa8, (byte)0xb3, (byte)0xba, (byte)0x85, (byte)0x8c, (byte)0x97, (byte)0x9e, (byte)0xe9, (byte)0xe0, (byte)0xfb, (byte)0xf2, (byte)0xcd, (byte)0xc4, (byte)0xdf, (byte)0xd6,
                    0x31, 0x38, 0x23, 0x2a, 0x15, 0x1c, 0x07, 0x0e, 0x79, 0x70, 0x6b, 0x62, 0x5d, 0x54, 0x4f, 0x46};

            //смешение столбцов
            private static byte[][] invMixColumns(byte[][] state) {
                byte[] temp = new byte[4];
                int i, j;

                for (i = 0; i <= NB - 1; i++) {
                    for (j = 0; j <= 3; j++)
                        temp[j] = state[j][i];
                    state[0][i] = (byte)(X_MUL_E[temp[0] & 0xFF] ^ X_MUL_B[temp[1] & 0xFF] ^ X_MUL_D[temp[2] & 0xFF] ^ X_MUL_9[temp[3] & 0xFF]);
                    state[1][i] = (byte)(X_MUL_9[temp[0] & 0xFF] ^ X_MUL_E[temp[1] & 0xFF] ^ X_MUL_B[temp[2] & 0xFF] ^ X_MUL_D[temp[3] & 0xFF]);
                    state[2][i] = (byte)(X_MUL_D[temp[0] & 0xFF] ^ X_MUL_9[temp[1] & 0xFF] ^ X_MUL_E[temp[2] & 0xFF] ^ X_MUL_B[temp[3] & 0xFF]);
                    state[3][i] = (byte)(X_MUL_B[temp[0] & 0xFF] ^ X_MUL_D[temp[1] & 0xFF] ^ X_MUL_9[temp[2] & 0xFF] ^ X_MUL_E[temp[3] & 0xFF]);
                }

                return state;
            }
        }
    }


}
